//--------------------------------------------------------------------------//
// xmtnimage1.cc                                                            //
//                                                                          //
// General (platform-independent) image display routines                    //
// Data conversion (itoa, gcvt)                                             //
// Program data initialization                                              //
// Main switch statement                                                    //
// Basic graphics (box, circle, line, etc.)                                 //
//                                                                          //
// Latest revision: 05-14-1999                                              //
// Copyright (C) 1999 by Thomas J. Nelson                                   //
// All rights reserved.                                                     //
//--------------------------------------------------------------------------//

#include "xmtnimage.h"
#include "xmtnimagec.h"

extern int         ci;                 // current image
extern Images     *z;
extern Globals     g;
extern PrinterStruct printer;
const int LABEL_LENGTH = 128;
const int COMMAND_LENGTH = 128;


//--------------------------------------------------------------------------//
// init_data                                                                //
// initialize global image arrays and other data                            //
// The 2d arrays are allocated as a 1d contiguous array, then variable-     //
//   aliased                                                                //
//--------------------------------------------------------------------------//
void init_data(void)
{
  if(g.diagnose){ printf("Initializing image data...");fflush(stdout); }
  const int CHARX=200, CHARY=200;  
  int j,k, size;
  int bytesperpixel;
  z = new Images[MAXIMAGES];
  ci=0;                                // Current image
  g.image_count = 0;                   // Total no. of images

  bytesperpixel = (7+g.bitsperpixel)/8;

  ////  Allocate stringbufr - max. CHARX x CHARY

  g.stringbufr_1d = (uchar*)malloc(CHARX*CHARY*sizeof(uchar)*bytesperpixel);
  if(g.stringbufr_1d==0) { fprintf(stderr,"%s for stringbufr_1d\n",g.nomemory); exit(1); }
  g.stringbufr    = make_2d_alias(g.stringbufr_1d, CHARX*bytesperpixel, CHARY);
  g.stringbufr_ximage = createximage(CHARX,CHARY,g.bitsperpixel,g.stringbufr_1d);
  memset(g.stringbufr_1d, g.bcolor, CHARX*CHARY*g.off[g.bitsperpixel]);
  g.spixmap = createxpixmap(g.main_window,CHARX,CHARY,g.bitsperpixel);
  for(k=0;k<INFOWINDOWS;k++)
      g.mpixmap[k] = createxpixmap(g.info_window[k],CHARX,CHARY,g.bitsperpixel);

  g.region = NULL;
  g.region_1d = NULL;
  g.macrotext = new char[MACROLENGTH];
  g.macrotext[0] = 0;

  size = (int)sqrt(g.xres*g.xres + g.yres*g.yres);  // Maximum diagonal distance
  g.highest = new int[2*size];   // For Bezier curves
  g.scan = new double[2*size];   // For densitometry
  
  ////  Crash if any of the above runs out of memory
  if(g.macrotext==0){ fprintf(stderr, g.nomemory); exit(1); }
  
  ////  After create_drawing_area(), it is safe to set the palette
  ////  and create XColormaps. 

  g.state = NORMAL;
  if(g.diagnose)
  {   printf("done\n");
      printf("cwd  %s\n",g.startdir);
      printf("size %d\n",size);
  }  
  ////  Make background total x,y size to avoid extra resizing.
  newimage(0,0,g.total_xsize,g.total_ysize,g.bitsperpixel,g.colortype,1,0,PERM,1,1,0);
  ////  Put background color in the background framebuffer.

  for(j=0;j<z[0].ysize;j++)  
      putpixelbytes(z[0].image[0][j],g.bcolor,z[0].xsize,g.bitsperpixel,1);

  if(g.bitsperpixel==8 && g.want_colormaps)  
  {    g.xcolormap = XCreateColormap(g.display, g.main_window, g.visual, AllocAll);  
       setpalette(g.palette);
  }


  repair(0);
  setimagetitle(0,"Background");
  if(g.autoundo) backupimage(0,0);
#ifdef HAVE_LEX
  initialize_evaluator();
#endif
}


//--------------------------------------------------------------------------//
// initialize_globals                                                       //
// initialize global data                                                   //
//--------------------------------------------------------------------------//
void initialize_globals(void)
{

    int i,k,len;
    char *ptr=NULL;
    if(g.diagnose){ printf("Initializing globals...");fflush(stdout); }
    g.dialogdone=0;
    g.main_window=0;                         // Application's window
    for(k=0;k<INFOWINDOWS;k++)
       g.info_window[k]=0;                   // The window at left for information
    g.mainmenuheight = 20;
    g.drawarea2width = 120;

    g.tif_xsize=1;
    g.tif_ysize=1;
    strcpy(g.version, "3.0.3");
    strcpy(g.emailaddress,"tjnelson@las1.ninds.nih.gov");
    strcpy(g.nomemory,"Insufficient memory");
    strcpy(g.scan_device, "/dev/scanner");  // Scanner device name
    strcpy(g.camera, "");         // Camera driver
    g.camera_id = 0;              // Which camera to use

    g.moving = 0;
    g.want_slew = 0;
    g.want_raise = 1;
    g.want_switching = 1;         // 1=clicking on image switches to it
    g.want_byte_order = -1;       // -1 = use default
    g.total_xsize = 540;          // X size of main window + information window
    g.total_ysize = 717;          // Y size of main window + information window
    g.main_xsize = g.total_xsize-g.drawarea2width;      // X size of main window
    g.main_ysize = g.total_ysize-(int)g.mainmenuheight; // Y size of main window
    g.main_xpos  = 0;             // Upper left corner of main window
    g.main_ypos  = 0;             // Upper left corner of main window
    g.getout=0;                   // Flag for 'cancel'

    g.want_colormaps=0;           // 1=want to set palette 2=not enough colors last time
    g.bitsperpixel=0;
    g.sparse_packing=0;           // 1=X windows sparse pseudo 24bpp mode
    g.autoundo=1;
    g.imode=1;
    g.want_messages=2;
    g.draw_figure=NONE;
    g.sprayfac=12;
    g.diameter=8;

    g.line.window=0;
    g.line.type=0;
    g.line.color=g.fcolor;
    g.line.width=1;
    g.line.wantgrad=0;
    g.line.grad=RGBvalue(128,128,128,24);
    g.line.skew=0;
    g.line.perp_ends=1;
    g.line.wantarrow=0;
    g.line.arrow_width=8;
    g.line.arrow_outer_length=12;
    g.line.arrow_inner_length=8;

    g.wantr=1;
    g.wantg=1;
    g.wantb=1;
    g.mouse_button=0;
    g.mouse_x=0;
    g.mouse_y=0;
    g.key_ascii=0;
    g.key_alt=0;
    g.key_ctrl=0;
    g.key_shift=0;
    g.image_count = 0;            // Total no. of images present
    g.imageatcursor=0;            // Image no. of pixel at cursor pos.
    g.text_direction=HORIZONTAL;
    g.want_format=-1;             // Overrides default save image format unless -1.
                                  // (Allows macro to specify file format on a 
                                  // separate line)

    g.read_cmyk=1;                // ==1 if image is to be interpreted as a cmyk
                                  // ==0 if upper 8 bits are to be thrown away
    g.save_cmyk=0;                // ==1 if create cmyk image
    g.usegrayscale=-1;
    g.want_quantize=1;            // ==1 if quantize,2=fit current palette 0=none
    g.jpeg_qfac=80;
    g.useheader=1;                // for custom image files
    g.xres=0;
    g.yres=0;
    g.signif = 4;                 // Significant digits to display 
    g.state=INITIALIZING;  // Program button-pressing state
    g.inmenu=0;                   // If in a menu, dialog box, sample palette,
                                  // etc. prevents putting pixels in
                                  // the current image buffer.
    g.false_color=0;              // no false color map
    g.tif_positive=1;
    g.tif_xoffset=0;
    g.tif_yoffset=0;
    g.tif_skipbytes=0;            // No. of extra junk bytes to skip at start
    g.bkgpalette=1;               // Palette of the background
    g.background_touched=0;       // Flag if any pixels were set on background
    g.csize=8;                    // Amount of cursor movement
    g.selected_ulx=0;
    g.selected_uly=0;
    g.selected_lrx=0;
    g.selected_lry=0;
    g.get_x1=0;
    g.get_y1=0;
    g.get_x2=0;
    g.get_y2=0;
    g.region_ulx=0;
    g.region_uly=0;
    g.region_lrx=0;
    g.region_lry=0;
    g.region_ino = -1;
    g.scancount=0;
    g.colortype=2;
    g.want_color_type=2;
    g.want_bpp=8;
    g.currentarrow = 0;
    g.selectedimage = 0;          // Most recently-selected image 
    g.selected_is_square = 1;     // 1 = selected area is rectangular
    g.want_shell=0;               // If 1 = each image in separate frame
    g.window_border=1;            // If 1 = windows get complete frame   
    g.wm_topborderwidth=0;        // Top border provided by window manager
    g.wm_leftborderwidth=0;       // Left border provided by window manager
    g.xlib_cursor = XC_crosshair; // Default cursor type
    g.fcolor   = 0;
    g.bcolor   = 192;
    g.fc.red   = 0;               // The desired foreground and background
    g.fc.green = 0;               //  rgb values. Used in selecting a new fcolor
    g.fc.blue  = 0;               //  and bcolor if g.fcolor or g.bcolor is already
    g.bc.red   = 32;              //  taken. Otherwise, the colors are determined
    g.bc.green = 32;              //  entirely by g.fcolor, g.bcolor, and g.b_palette.
    g.bc.blue  = 32;
    g.luminr = 0.299;
    g.luming = 0.587;
    g.luminb = 0.114;
    g.font = new char[256];
    strcpy(g.font,"*helvetica-bold-r-normal--12*");
    g.imagefont = new char[256];
    strcpy(g.imagefont,"*adobe-times-bold-r-normal--24-240-75-75*");
    g.window_handle_size = 2;     // Width at edge where clicking moves image
    g.nbuttons=0;
    g.ignore=0;
    g.spacing=4;
    g.text_spacing = 0;
    g.object_threshold.red = 50;  // 0-255
    g.object_threshold.green = 50;// 0-255
    g.object_threshold.blue = 50; // 0-255
    g.double_click_msec=400;      // milliseconds to wait for double clicking
    g.alpha=1;                    // 1=alpha channel is active
    g.nbuttons=0;
    g.zoom_button=-1;             // Which button at left is for zoom
    g.copy=0;                     // 0=move 1=copy
    g.want_crosshairs=0;          // 1=full-screen crosshairs
 
    g.openlistcount = 0;
    g.openlist = new listinfo*[100];  // Array for pointers to up to 100 lists
                                      // that want to be dynamically updated
    if(g.openlist==NULL) { printf("Insufficient memory\n"); exit(1); }
                                         
    for (i=0; i<64; i++)              // Define grayscale colormap
    {
            g.gray_palette[i*4].red      =  i;      //black to white
            g.gray_palette[i*4].green    =  i;
            g.gray_palette[i*4].blue     =  i;

            g.gray_palette[i*4+1].red      =  i;
            g.gray_palette[i*4+1].green    =  i;
            g.gray_palette[i*4+1].blue     =  min(63,i+1);

            g.gray_palette[i*4+2].red      =  i;
            g.gray_palette[i*4+2].green    =  min(63,i+1);
            g.gray_palette[i*4+2].blue     =  i;

            g.gray_palette[i*4+3].red      =  i;
            g.gray_palette[i*4+3].green    =  min(63,i+1);
            g.gray_palette[i*4+3].blue     =  min(63,i+1);           
    }
    g.gray_palette[0].red = 0;
    g.gray_palette[0].green = 0;
    g.gray_palette[0].blue = 0;
    memcpy(g.fc_palette, g.gray_palette,768);
 
    ////  Paths
  
    ////  Current directory
    getcwd(g.currentdir, FILENAMELENGTH-1);  
    len = strlen(g.currentdir)-1;            // eliminate trailing '/'
    if(g.currentdir[len]=='/') g.currentdir[len]=0;

    ////  Start-up dir.(return here before exit); location for print temp files
    strcpy(g.startdir, g.currentdir);        

    ////  $HOME/.tnimage for tnimage.ini, formats in Bourne
    ////  $home/.tnimage in tcsh
    ptr = getenv("HOME");
    if(ptr!=NULL) strcpy(g.homedir, ptr);    
    else 
    {   ptr = getenv("home");
        if(ptr!=NULL) strcpy(g.homedir, ptr);    
        else 
        {   strcpy(g.homedir, g.startdir);
            fprintf(stderr,"Unable to determine home directory in __FILE__ line __LINE__\n");
        }
    }
    strcat(g.homedir,"/.tnimage");

    ////  Location of tnimage.hlp, etc.
    strcpy(g.helpdir, HELPDIR);   
    sprintf(g.formatpath,"%s/%s",g.homedir,"formats");
    sprintf(g.waveletpath,"/usr/local/lib/wavelets");

    strcpy(g.compression,"/bin/gzip -f");       // Program to exec to compress image
    strcpy(g.decompression, "/bin/gunzip -c");  // Program to exec to decompress image
    strcpy(g.compression_ext,".gz");            // Extension added by compression program

    ////  Initialize printer defaults

    printer.copies = 1;
    printer.hsize = 4.0;
    printer.vsize = 11.0;
    printer.vpos = 0.0;
    printer.entire = 1;
    printer.hpos = 0.0;
    printer.ratio = 1.0;
    printer.intensity = 1;
    printer.paperxsize = 8.5;
    printer.paperysize = 11.0;
    printer.ifac = 1.0;
    printer.rfac = 1.0;
    printer.gfac = 1.0;
    printer.bfac = 1.0;
    printer.papertype = 0;
    printer.graybalance = 0;
    printer.depletion = 0;
    strcpy(printer.command, "lpr -P");
    strcpy(printer.devicename, "/dev/lp2");
    strcpy(printer.filename, "temp.print");
    printer.vertical = 1;
    printer.positive = 1;
    printer.palette = 2;         // 1-4 for B/W, RGB,CMY, or CMYK
    printer.dither = 8;          // # of printer pixels/screen pixel
    printer.quality = 0;
    printer.res = 300;
    printer.rotation = 0;
    printer.interpolate = 0;
    printer.language=0;          // 0=postscript 1=pcl
    printer.device = 1;


    if(g.diagnose)
    {   printf("Done\n");
        printf(" Version %s\n",g.version);
        fflush(stdout); 
    }

}


//--------------------------------------------------------------------------//
// execute                                                                  //
// execute a command from menu or a macro. This method is better than       //
// having a separate callback for each menu item because it makes it        //
// easy to use macros.                                                      //
//--------------------------------------------------------------------------//
int execute(char *command, int noofargs, char **arg)
{
  static char **oldarg=NULL;                // up to 20 arguments
  static char oldo[256];
  static int oldnoofargs=0;
  int ino,f,j,k,status=OK;
  char o[256], tempstring[256];
  g.draw_figure=NONE;
  g.state=NORMAL;
  g.getout=0;
  
  int k2;
  int e=0;   // flag if something was executed

  if(oldarg==NULL)                          // remain allocated forever
  {     oldarg = new char*[20];
        for(k=0;k<20;k++){ oldarg[k] = new char[128]; oldarg[k][0]=0; }
  }
  if(g.diagnose){ printf("Executing %s, %d args\n",command,noofargs); fflush(stdout);}
  if(g.inmenu<0) fprintf(stderr,"Error: inmenu=%d\n",g.inmenu);
  for(k=0,j=0;k<=(int)strlen(command);k++)
  {  switch(command[k])
       {   case '.':
           case '_':
           case ' ':
           case '/': break;
           default : o[j++] = tolower(command[k]);
                     o[j]=0;
       }
  }

  //----------- The commands (in alphabetical order) -------------//
  if(strcmp(o, "repeat")==SAME && strcmp(o, "undo")!=SAME)
  {     if(strlen(oldo)) strcpy(o, oldo);
        noofargs = oldnoofargs;
        for(k=0;k<noofargs;k++) strcpy(arg[k], oldarg[k]);
  }
  if(strcmp(o, "3d")==SAME){ e=1; menu3d(); }
  if(strcmp(o, "3dplot")==SAME){ e=1; plot3d(ci); }
  if(strcmp(o, "aboutthefile")==SAME){ e=1; aboutthefile(); }
  if(strcmp(o, "abouttheimage")==SAME){ e=1; abouttheimage(); }
  if(strcmp(o, "abouttheprogram")==SAME){ e=1; abouttheprogram(); }
  if(strcmp(o, "acquire")==SAME){ e=1; acquire(); }
  if(strcmp(o, "addborder") == SAME){ e=1; image_border(); }
  if(strcmp(o, "attributes")==SAME){ e=1; attributes(); } 
  if(strcmp(o, "automaticundo")==SAME){ e=1; g.autoundo=atoi(arg[1]); }
  if(strcmp(o, "backgroundcolor")==SAME){ e=1; setbcolor(); }
  if(strcmp(o, "backup")==SAME){ e=1; backupimage(ci,1); }
  if(strcmp(o, "beep") == SAME){ e=1; XBell(g.display,0);  }
  if(strcmp(o, "border") == SAME){ e=1; image_border(); }
  if(strcmp(o, "box")==SAME)
  {  e=1; 
     g.line.width = clickbox("Box width",g.line.width,1,100,null,0);
     g.draw_figure=BOX; 
  }
  if(strcmp(o, "brightness")==SAME){ e=1; changebrightness(noofargs,arg); }
  if(strcmp(o, "calibration")==SAME){ e=1; calibrate();  }
  if(strcmp(o, "camera")==SAME){ e=1; camera(); }
  if(strcmp(o, "cancel")==SAME)
  {  e=1; 
     maincancelcb(0,0,0); 
     unpush_null_buttons();
     return ABORT;      //// Break out so g.getout isnt reset
  }
  if(strcmp(o, "changecolordepth")==SAME){ e=1;status=changecolordepth(noofargs,arg);}
  if(strcmp(o, "changeimagedepth")==SAME){ e=1;status=changecolordepth(noofargs,arg);}
  if(strcmp(o, "changecolormap")==SAME){ e=1; changepalette(noofargs,arg); }
  if(strcmp(o, "changecontrast")==SAME){ e=1; changecontrast(noofargs,arg); }
  if(strcmp(o, "changepixelvalues")==SAME){ e=1; addvalue(noofargs,arg);    } 
  if(strcmp(o, "changesize")==SAME){ e=1; status=shrink(noofargs,arg); }
  if(strcmp(o, "changetitle")==SAME || strcmp(o, "title")==SAME){ e=1; changetitle(noofargs,arg);}
  if(strcmp(o, "circle")==SAME)
  {   e=1;
      g.diameter = clickbox("Diameter",g.diameter,0,400,null,0);
      g.draw_figure=CIRCLE;
  }
  if(strcmp(o, "clearalpha")==SAME){ e=1; clear_alpha(ci); }
  if(strcmp(o, "clearalphachannel")==SAME){ e=1; clear_alpha(ci); }
  if(strcmp(o, "color")==SAME){ e=1; changebrightness(noofargs,arg); }
  if(strcmp(o, "color->grayscale")==SAME){ e=1; converttograyscale(ci); }
  if(strcmp(o, "colorgrayscale")==SAME){ e=1; converttograyscale(ci); }
  if(strcmp(o, "colorbrightness")==SAME){ e=1; changebrightness(noofargs,arg); }
  if(strcmp(o, "colormap")==SAME){ e=1; changepalette(noofargs,arg); }
  if(strcmp(o, "compositergb")==SAME){ e=1;status=combine_colors(noofargs,arg); }
  if(strcmp(o, "configure")==SAME){ e=1; configure(); }
  if(strcmp(o, "contrast")==SAME){ e=1; changecontrast2(noofargs,arg); }
  if(strcmp(o, "convert")==SAME){ e=1; status=changecolordepth(noofargs,arg); }
  if(strcmp(o, "createfileformat")==SAME){ e=1; create_custom_format(); }
  if(strcmp(o, "createimage")==SAME){ e=1; status=createimage(noofargs,arg); }
  if(strcmp(o, "createresizeimage")==SAME){ e=1; status=createimage(noofargs,arg); }
  if(strcmp(o, "crop")==SAME){ e=1; crop(); }
  if(strcmp(o, "curve")==SAME){ e=1; curve(); } 
  if(strcmp(o, "darken")==SAME){ e=1; darken(noofargs, arg); }
  if(strcmp(o, "deleteregion")==SAME){ e=1; paint(g.bcolor); }
  if(strcmp(o, "down") == SAME){ e=1; downcb(0,0,0);  g.moving=0;}
  if(strcmp(o, "duplicate")==SAME){ e=1; duplicate(noofargs, arg); }
  if(strcmp(o, "erasebackground")==SAME){ e=1; erasebackground(); }
  if(strcmp(o, "erasefft") == SAME){ e=1; erasefft(ci); }
  if(strcmp(o, "evaluate") == SAME){ e=1; evaluate(noofargs,arg); }
  if(strcmp(o, "executeplugin") == SAME){ e=1; select_plugin(noofargs,arg); }
  if(strcmp(o, "exit") == SAME){ e=1; status=QUIT; }
  if(strcmp(o, "fft")==SAME){ e=1; status=transform(noofargs,arg); }
  if(strcmp(o, "fftdisplay")==SAME)
  {    e=1; 
       if(ci>=0)
       {    z[ci].fftdisplay=atoi(arg[1]); 
            scalefft(ci);
            rebuild_display(ci);  
            redraw(ci);
       }
  }
  if(strcmp(o, "fileformat")==SAME){ e=1; g.want_format=atoi(arg[1]); }
  if(strcmp(o, "fill")==SAME){ e=1;  fillregion();  }
  if(strcmp(o, "fillregion")==SAME){ e=1;  fillregion(); }
  if(strcmp(o, "filter")==SAME){ e=1; filter(noofargs,arg);  }
  if(strcmp(o, "fliphoriz")==SAME){ e=1; flip_horiz(); }     
  if(strcmp(o, "fliphorizontally")==SAME){ e=1; flip_horiz(); }     
  if(strcmp(o, "flipvertically")==SAME){ e=1; flip_vert(); }
  if(strcmp(o, "font")==SAME){ e=1; changefont(); }
  if(strcmp(o, "foregroundcolor")==SAME){ e=1; setfcolor(); }
  if(strcmp(o, "frame")==SAME)
  {    e=1; 
       f=atoi(arg[1]);
       if(ci>=0 && between(f,0,z[ci].frames-1))
       {    z[ci].cf = f; 
            rebuild_display(ci); 
            redraw(ci);
       }
  }
  if(strcmp(o, "grayscale->color")==SAME || strcmp(o, "grayscalecolor")==SAME)
  {   e=1; converttocolor(ci); }
  if(strcmp(o, "grains")==SAME) { e=1; pattern_recognition(); }
  if(strcmp(o, "grayscalemap")==SAME){ e=1; grayscalemap(); }
  if(strcmp(o, "help") == SAME){ e=1; help(1); }
  if(strcmp(o, "histogram")==SAME){ e=1; histogram(); }
  if(strcmp(o, "image") == SAME){ strcpy(o,"selectimage"); }
  if(strcmp(o, "imagemath")==SAME){ e=1; image_math(); }
  if(strcmp(o, "mask")==SAME){ e=1; mask(noofargs,arg); } 
  if(strcmp(o, "math")==SAME){ e=1; image_math(); } 
  if(strcmp(o, "maximize") == SAME){ e=1; maximize_contrast(noofargs,arg); }
  if(strcmp(o, "maximizecontrast") == SAME){ e=1; maximize_contrast(noofargs,arg); }
  if(strcmp(o, "intensity")==SAME){ e=1; change_intensity(noofargs,arg); }
  if(strcmp(o, "invertcolors")==SAME){ e=1; invertcolors(); }
  if(strcmp(o, "label")==SAME){ e=1; label(); }
  if(strcmp(o, "left") == SAME){ e=1; leftcb(0,0,0);  g.moving=0;}
  if(strcmp(o, "lighten")==SAME){ e=1; lighten(noofargs, arg); }
  if(strcmp(o, "line")==SAME || strcmp(o, "arrow")==SAME)
  {  e=1; 
     set_line_parameters();
     if(!g.getout)
     {   if(g.line.wantarrow)
             g.draw_figure=ARROW;
         else
             g.draw_figure=LINE; 
     }
  }
  if(strcmp(o, "linestyle")==SAME){ e=1; set_line_parameters(); }
  if(strcmp(o, "readfft") == SAME){ e=1; readfft(noofargs,arg); }
  if(strcmp(o, "loadfft") == SAME){ e=1; readfft(noofargs,arg); }
  if((strcmp(o,"loadimage")==SAME)||(strcmp(o,"load")==SAME)||(strcmp(o,"open")==SAME)||
     (strcmp(o,"read")==SAME))
  {  e=1; 
     int count=0, status1=0;
     char **filenames=NULL;
     filenames = setreadparameters(&status, &count, noofargs, arg);
     if(!g.getout && filenames !=NULL) 
     {   for(k=0;k<count;k++)
         { 
             status1 = readimage(filenames[k],noofargs,arg);
             status = max(status, status1);
         }
     }
     for(k=0;k<count;k++) XtFree(filenames[k]);
     if(count) delete[] filenames;
     g.getout=0;
  }
  if(strcmp(o, "macro")==SAME)
  {  e=1;
     if(noofargs>0) status=macro(g.macrotext);
     else  
     {  edit(g.appname,
             "Command editor (Shift-Enter to execute command)",
             g.macrotext,80,50,MACROLENGTH,30);   
        status=macro(g.macrotext); 
     } 
  }    
  if(strcmp(o, "measure")==SAME){ e=1; measure(); }
  if(strcmp(o, "messages")==SAME)
  {   e=1; 
      if(noofargs>0) g.want_messages = atoi(arg[1]); 
  }
  if(strcmp(o, "move")==SAME){ e=1; g.copy=1-g.copy; reset_user_buttons(); }
  if(strcmp(o, "movie") == SAME){ e=1; movie(atoi(arg[1])); }
  if(strcmp(o, "moveto")==SAME)
  {   e=1;
      if(noofargs>0) z[ci].xpos = atoi(arg[1]); 
      if(noofargs>1) z[ci].ypos = atoi(arg[2]); 
      moveimage(ci, z[ci].xpos, z[ci].ypos);
  }
  if(strcmp(o, "new")==SAME){ e=1; status=createimage(noofargs,arg); }
  if(strcmp(o, "null")==SAME)
  {  e=1; 
     unpush_null_buttons(); 
     return ABORT;      //// Break out so g.getout isnt reset
  }
  if(strcmp(o, "paint")==SAME){ e=1; paint(g.fcolor); }
  if(strcmp(o, "paintregion")==SAME){ e=1; paint(g.fcolor); }
  if(strcmp(o, "palette")==SAME){ e=1; changepalette(noofargs,arg); }
  if(strcmp(o, "partition")==SAME){ e=1; partition(); }
  if(strcmp(o, "paste")==SAME){ e=1; pasteimage(noofargs,arg); }
  if(strcmp(o, "patternrecognition")==SAME) { e=1; pattern_recognition(); }
  if(strcmp(o, "pixelinteractmode")==SAME){ e=1; pixelmode(noofargs,arg); }
  if(strcmp(o, "printimage") == SAME){ e=1; status=printimage(); }
  if(strcmp(o, "print") == SAME){ e=1; status=printimage(); }
  if(strcmp(o, "shiftframe") == SAME){ e=1; shift_frame(noofargs,arg); }
  if(strcmp(o, "shift") == SAME){ e=1; shift_frame(noofargs,arg); }
  if(strcmp(o, "quit") == SAME) { e=1; quitcb(0,0,0);; }          
  if(strcmp(o, "remapcolors") == SAME){ e=1; remapcolors(); }
  if(strcmp(o, "resize")==SAME){ e=1; resize(noofargs, arg); }
  if(strcmp(o, "resizewindow")==SAME){ e=1; resize(noofargs, arg); }
  if(strcmp(o, "resizeimage")==SAME){ e=1; status=createimage(noofargs,arg); }
  if(strcmp(o, "redraw") == SAME){ e=1; redraw(ci); } 
  if(strcmp(o, "repair") == SAME){ e=1; repair(ci); } 
  if(strcmp(o, "restore") == SAME){ e=1; restoreimage(); } 
  if(strcmp(o, "right") == SAME){ e=1; rightcb(0,0,0); g.moving=0; }
  if(strcmp(o, "root") == SAME){ e=1; root(ci, ON); } 
  if(strcmp(o, "rootpermanent") == SAME){ e=1; root(ci, PERMANENT_TILE); } 
  if(strcmp(o, "rootoff") == SAME){ e=1; root(ci, OFF); } 
  if(strcmp(o, "rotate")==SAME){ e=1;  status=rotate(noofargs,arg); } 
  if(strcmp(o, "rotateimage")==SAME){ e=1; status=rotate(noofargs,arg); }
  if(strcmp(o, "save")==SAME){ e=1; status=saveimage(ci,noofargs,arg); }
  if(strcmp(o, "savefft") == SAME){ e=1; writefft(ci); }
  if(strcmp(o, "saveimage")==SAME){ e=1;status=saveimage(ci,noofargs,arg); }
  if(strcmp(o, "scan")==SAME){ e=1; acquire(); }
  if( ((strcmp(o, "selectaregion")) == SAME)||
      ((strcmp(o, "selectregion")) == SAME) ||
      ((strcmp(o, "select")) == SAME))
  {  e=1;
     if(noofargs==0)
        getbox(g.selected_ulx, g.selected_uly, g.selected_lrx, g.selected_lry); 
     else
        select_region(atoi(arg[1]),atoi(arg[2]),atoi(arg[3]),atoi(arg[4]));
  }
  if(strcmp(o, "selectirregulararea")==SAME ||
     strcmp(o, "selectarea")==SAME || 
     strcmp(o, "manuallyselectarea")==SAME ||
     strcmp(o, "scissors")==SAME) { e=1; select_area(MULTIPLE); }
  if(strcmp(o, "scanner")==SAME){ e=1; acquire(); }
  if((strcmp(o, "selectimage") == SAME) ||
     (strcmp(o, "switchto") == SAME))
  {  e=1;
     if(noofargs>=1)
     {    ino = image_number(arg[1]);
          switchto(ino); 
     }else image_list();
  }   
  if(strcmp(o, "separatergb")==SAME){ e=1; status=separate_colors(ci,noofargs,arg); }
  if(strcmp(o, "showalpha")==SAME){ e=1; show_alpha(g.selectedimage); }
  if(strcmp(o, "showalphachannel")==SAME){ e=1; show_alpha(g.selectedimage); }
  if(strcmp(o, "showodtable")==SAME){ e=1; showodtable(); }
  if(strcmp(o, "showpalette")==SAME){ e=1; samplepalette(); } 
  if(strcmp(o, "showcolormap")==SAME){ e=1; samplepalette(); } 
  if(strcmp(o, "showselected")==SAME){ e=1; show_selected(g.selectedimage); }
  if(strcmp(o, "size")==SAME){ e=1; status=shrink(noofargs,arg); }
  if(strcmp(o, "sketch")==SAME)
  {   e=1; 
      if(g.draw_figure==SKETCH) g.draw_figure=NONE; 
      else g.draw_figure=SKETCH; 
      reset_user_buttons();  
  }
  if(strcmp(o, "slew")==SAME){ e=1; g.want_slew=1-g.want_slew; reset_user_buttons(); }
  if(strcmp(o, "spotdensitometry")==SAME){ e=1; densitometry(); }
  if(strcmp(o, "spray")==SAME) { e=1; status=spray(noofargs,arg); }
#ifdef HAVE_XBAE
  if(strcmp(o, "spreadsheet")==SAME) { e=1; create_spreadsheet(ci); }
#endif
  if(strcmp(o, "step") == SAME)
  {   e=1; 
      if(noofargs==0) g.csize = clickbox("Step size",g.csize,0,1000,null,0);
      else g.csize = atoi(arg[1]);
  }
  if(strcmp(o, "stripdensitometry")==SAME){ e=1; scan_region(); }
  if(strcmp(o, "switchselected")==SAME){ e=1; switch_selected_region(); }
  if(strcmp(o, "testplugin") == SAME)
  {    e=1; 
       strcpy(arg[2],"0");
       select_plugin(noofargs,arg); 
  }
  if(strcmp(o, "textdirection")==SAME)
  {  e=1;
     if(g.text_direction==HORIZONTAL) g.text_direction=VERTICAL;
     else g.text_direction=HORIZONTAL;
     message("Text direction changed");
  }
  if(strcmp(o, "localcontrast")==SAME){ e=1; localcontrast(noofargs, arg); }
  if(strcmp(o, "tracecurve")==SAME){ e=1; status=trace_curve(); }
  if(strcmp(o, "chromakey")==SAME)
  {   e=1; 
      if(ci>=0) 
      {   z[ci].chromakey = 1-z[ci].chromakey; 
          rebuild_display(ci);
          redraw(ci);
          redraw_chromakey_image(ci);
      }
  }
  if(strcmp(o, "erase")==SAME)
  {   e=1;
      message("Spray erase mode\nClick Cancel when finished");
      sprintf(arg[1],"%d",ERASESPRAY);
      noofargs=1;
      spray(noofargs,arg);
  }
  if(strcmp(o, "transparency")==SAME){ e=1; set_transparency(ci,noofargs,arg); }
  if(strcmp(o, "undo") == SAME){ e=1; restoreimage(); } 
  if(strcmp(o, "unloadimage") == SAME ||
     strcmp(o, "unload") == SAME      ||
     strcmp(o, "eraseimage") == SAME       ||
     strcmp(o, "close") == SAME)  { e=1; unloadimage(noofargs,arg); } 
  if(strcmp(o, "unloadall") == SAME) 
  {    e=1;
       k2 = g.image_count;
       //// This changes g.image_count and renumbers the images
       for(k=1; k<k2; k++) eraseimage(1,1,0,0); 
       redrawscreen();
  }
  if(strcmp(o, "unzoom")==SAME){ e=1; repair(ci); }
  if(strcmp(o, "up") == SAME){ e=1; upcb(0,0,0); g.moving=0;}
  if(strcmp(o, "wallpaper") == SAME){ e=1; for(k=0;k<2;k++) root(ci, PERMANENT_TILE); } 
  if(strcmp(o, "warp")==SAME){ e=1; warp(); }
  if(strcmp(o, "wavelets")==SAME){ e=1; wavelet(noofargs,arg); }
  if(strcmp(o, "whatis") == SAME){ e=1; evaluate(noofargs,arg); }
  if(strcmp(o, "windows")==SAME){ e=1; g.want_shell=atoi(arg[1]); }
  if(strcmp(o, "windowborder")==SAME){ e=1; g.window_border=atoi(arg[1]); }
  if(strcmp(o, "zoom")==SAME){ e=1; zoom(ci); }

  //------------------------------------------------//

  g.getout=0; 
  if(e==0) status=UNKNOWN;
  g.state=NORMAL;
  if(g.draw_figure) g.state=DRAWING;
  printstatus(g.state);
  if(g.diagnose)
  {  printf(" status=%d, %d, %d, %d\n",status,g.state,g.draw_figure,g.inmenu); 
     printf(" images=%d ci=%d\n",g.image_count,ci); 
     print_image_info();
     fflush(stdout);
  }

  if(e)
  {   for(k=0;k<g.openlistcount;k++) update_list(k);
      for(k=0;k<g.image_count;k++) 
           if(z[k].s->visible && z[k].touched) putimageinspreadsheet(k);
      update_sample_palette();
  }else if(strlen(command))
  {   sprintf(tempstring,"%s ",command);
      for(k=1;k<=noofargs;k++)
      {   strcat(tempstring,arg[k]); 
          strcat(tempstring," "); 
      }
      math(tempstring);
  }
  if(g.autoundo==2 && ci>=0) backupimage(ci,0);
  strcpy(oldo, o);
  for(k=0;k<noofargs;k++) strcpy(oldarg[k], arg[k]);
  oldnoofargs = noofargs;
  return status;
}



//--------------------------------------------------------------------------//
// setimagetitle                                                            //
//--------------------------------------------------------------------------//
void setimagetitle(int ino, char *title)
{   
    if(ino>=0 && strlen(title))
    {    strcpy(z[ino].name, title);    
         if(z[ino].shell)
             XtVaSetValues(XtParent(XtParent(z[ino].widget)), XmNtitle,
                 z[ino].name,  XmNiconName, z[ino].name, 0);
    }
}


//--------------------------------------------------------------------------//
// set_transparency                                                         //
//--------------------------------------------------------------------------//
void set_transparency(int ino, int noofargs, char **arg)
{   
    if(ino<0) return;
    int t=z[ino].transparency;
    if(noofargs==0) t = clickbox("Transparency",t,0,100,null,43);
    else t = atoi(arg[1]);
    z[ino].transparency = t;
    rebuild_display(ino);
    redraw(ino);
    redraw_transparent_image(ino);
}



//--------------------------------------------------------------------------//
//copybackground  Copy parts of the image in corresponding                  //
//  screen locations to the screen.                                         //
//--------------------------------------------------------------------------//
void copybackground(int x1, int y1, int x2, int y2, int skipimage)
{
    int k,L,ytest,ino;
    int ix1,ix2,iy1,iy2,startx,endx,starty,endy;

    if(x1<0) x1=0;               // Fix the parameters  
    if(x2<0) x2=0;
    if(x1>g.main_xsize-1) x1=g.main_xsize-1;
    if(x2>g.main_xsize-1) x2=g.main_xsize-1;
    if(x2<x1)swap(x1,x2);
    ytest=0;
    if(y1<ytest) y1=ytest;
    if(y2<ytest) y2=ytest;
    if(y1>g.main_ysize-1)y1=g.main_ysize-1;
    if(y2>g.main_ysize-1)y2=g.main_ysize-1;
    if(y2<y1)swap(y1,y2);
                                 // Copy portions of any image in region
                                 // in order of back to front.
                                 // Only the portion between startx,starty
                                 // and endx, endy is copied.
    for(k=g.image_count-1; k>=0; k--)
    {   
          ino = z[k].overlap;
          if((ino>=0)&&(ino!=skipimage))
          {     ix1 = 0;         // Edges of image (can be off screen)
                ix2 = z[ino].xsize - 1; 
                iy1 = 0;
                iy2 = z[ino].ysize - 1;
                starty = max(iy1, y1-z[ino].ypos);
                endy   = min(iy2, y2-z[ino].ypos);
                if(endy>starty)
                {       startx = max(ix1, x1-z[ino].xpos-1); //Start, end x pos.on image
                        endx   = min(ix2, x2-z[ino].xpos-1);
                        L = endx - startx;                 // no.of x pixels to copy 
                        if(L>0)              
                          copyimage(startx,starty,endx,endy,startx-ix1,starty-iy1,
                              PUT, z[ino].img_ximage, z[ino].win);         
                }   
          }
    }       
}



//--------------------------------------------------------------------------//
// erasefromscreen - replace image with background on screen                //
//--------------------------------------------------------------------------//
void erasefromscreen(int ino)
{   
    int x1,x2,y1,y2;
    if(ino>=0)
    { x1 = z[ino].xpos;          // Edges of image (can be off screen)
      x2 = z[ino].xpos+z[ino].xsize;
      y1 = z[ino].ypos;
      y2 = z[ino].ypos+z[ino].ysize;
      copybackground(x1,y1,x2,y2,ino);
    }  

   // Temporarily move image off edge of the screen to prevent XOR from
   // reading its pixels. 

   z[ino].xpos = g.main_xsize + 1;
   z[ino].ypos = g.main_ysize + 1;

}

//--------------------------------------------------------------------------//
// switchto - switch screen to image no. ino                                //
//--------------------------------------------------------------------------//
void switchto(int ino)
{ 
  int j,k,level;
  if(!g.want_switching) return;
  if(ino>=0 && ino<g.image_count)
  {   drawselectbox(OFF);
      if(ino>0)
      {  if(ci!=ino && g.want_raise)
         {   
             if(z[ino].shell) 
                XRaiseWindow(g.display,XtWindow(XtParent(XtParent(z[ino].widget))));
             else
                XRaiseWindow(g.display,z[ino].win);
         }
         level = g.image_count -1;    // Re-sort images so ino is on top
         for(j=0; j<g.image_count; j++)
            if(z[j].overlap==ino) level=j;
         for(j=level;j>0;j--)
            z[j].overlap = z[j-1].overlap;
         z[0].overlap = ino;
      }
      ci = ino;                                
      if(z[ci].palette!=NULL)
      {  setpalette(z[ci].palette);
         g.currentpalette = -2;
      }
      selectimage(ci);
      for(k=0;k<g.openlistcount;k++) update_list(k);
      update_sample_palette();
      if(g.zoom_button!=-1 && g.button[g.zoom_button]->state != z[ci].is_zoomed)
      {   if(z[ci].is_zoomed) push_main_button(g.zoom_button);
          else unpush_main_button(g.zoom_button);
      }
  }
}      

 
//--------------------------------------------------------------------------//
// selectimage - select image # ino                                         //
//--------------------------------------------------------------------------//
void selectimage(int ino)
{
  drawselectbox(OFF);
  if(!between(ino,0,g.image_count-1)) return;
  g.selectedimage = ino; 
  g.selected_is_square=1;
  g.selected_ulx = z[ino].xpos;
  g.selected_lrx = z[ino].xpos + z[ino].xsize-1;
  g.selected_uly = z[ino].ypos;
  g.selected_lry = z[ino].ypos + z[ino].ysize-1;    
  g.ulx = zoom_x_coordinate(g.selected_ulx, ino);
  g.uly = zoom_y_coordinate(g.selected_uly, ino);
  g.lrx = zoom_x_coordinate(g.selected_lrx, ino);
  g.lry = zoom_y_coordinate(g.selected_lry, ino); 
}


//--------------------------------------------------------------------------//
// select_region                                                            //
//--------------------------------------------------------------------------//
void select_region(int x1, int y1, int x2, int y2)
{
    g.selected_ulx = x1;
    g.selected_uly = y1;
    g.selected_lrx = x2;
    g.selected_lry = y2;
    g.selectedimage=-1;
    g.ulx = x1;
    g.uly = y1;
    g.lrx = x2;
    g.lry = y2;
    g.selected_is_square=1;
    drawselectbox(ON);
}


//--------------------------------------------------------------------------//
// setpixel (Xlib) - Plot point to screen.                                  //
//  'mode' determines how pixel interacts with color at that location.      //
// Returns the actual color set (may be different from color if mode        //
//   is something other than SET).                                          //
// Default ino is -2 => automatically calculates ino from x,y               //
// Default win is  0 => Use main window                                     //
// win and ino are optional parameters.                                     //
//--------------------------------------------------------------------------//
uint setpixel(int x, int y, uint color, int mode, Window win, int ino)
{
    static int hit=0;
    int bpp,i,j;
    uchar *add;
    GC *gc;
    if(win==0) 
    {   win = g.main_window;
        gc = &g.image_gc;
        switch(mode)
        {   case SET: break;
            case MAXIMUM: color=max(color,readpixel(x,y)); break;
            case MINIMUM: color=min(color,readpixel(x,y)); break;     
            case ADD: color=min(g.maxcolor,readpixel(x,y)+color); break;
            case SUB12: color=(uint)max(0,(int)color-(int)readpixel(x,y));break;
            case SUB21: color=(uint)max(0,(int)readpixel(x,y)-(int)color);break;
            case XOR: color^=readpixel(x,y); break;
            case AVE: color+=readpixel(x,y); color>>=1; break;       
            case SUP: if(color==0) return 0; break;
            default:  break;
        }
    }else
    { 
        switch(mode)
        {   case XOR:  gc = &g.xor_gc; break;
            default:   gc = &g.image_gc; break; 
        }
    }

    if(color!=g.fcolor || !hit) XSetForeground(g.display, *gc, (ulong)color);
    hit=1;

    ////  If drawing on drawing_area or in an image, copy the pixel value
    ////  into the appropriate buffer.
    
    if(!g.inmenu && ino==-2) ino=whichimage(x,y,bpp);
    if(ino>=0 && z[ino].is_zoomed)
    {   
         x = zoom_x_screen(x,ino);
         y = zoom_y_screen(y,ino);
         if(x<0 || x>=z[ino].xsize || y<0 || y>=z[ino].ysize) return 0;
         for(j=0;j<=(int)(0.99+z[ino].zoomy);j++)
         for(i=0;i<=(int)(0.99+z[ino].zoomx);i++)
         {
              XDrawPoint(g.display, z[ino].win, g.image_gc, x+i, y+j);  
              if(x+i<0 || y+j<0 || x+i>=z[ino].xsize || y+j>=z[ino].ysize) continue;
              if(g.bitsperpixel==8)  // Branch for speed for different screen depths
                   z[ino].img[y+j][x+i] = color;
              else
              {    add = z[ino].img[y+j] + (x+i)*g.off[g.bitsperpixel];
                   putpixelbytes(add, color);
              }
         }
         
    }else if(win==g.main_window || ino>=0)
    {    if(ino==-2) ino=whichimage(x,y,bpp);
         if(ino>=0)
         {   x -= z[ino].xpos;
             y -= z[ino].ypos;

             ////  If setpixel() is called while image is moving, x & y can go 
             ////  out of bounds, so jump out.

             if(x<0 || x>=z[ino].xsize || y<0 || y>=z[ino].ysize) return 0;
         }else
             if(x<0 || x>=g.main_xsize || y<0 || y>=g.main_ysize) return 0;


         // Since each image is in a different window, subtract the image's
         // position from the coordinates.

         if(g.bitsperpixel==8)  // Branch for speed for different screen depths
         {  z[ino].img[y][x] = color;
         }else
         {  add = z[ino].img[y]+ x*g.off[g.bitsperpixel];
            putpixelbytes(add, color);
         }
         win = z[ino].win;
         XDrawPoint(g.display, win, *gc, x, y);  
    }else
    {
         XDrawPoint(g.display, win, *gc, x, y);  
    }
    if(color!=g.fcolor) XSetForeground(g.display, *gc, g.fcolor); 
    return color;
}
  
  


//--------------------------------------------------------------------------//
//setpixelonimage - Set color of a point in the selected current image      //
//  color    = desired color to use                                         //
//  bpp      = bits/pixel of `color' input value - not necessarily the      //
//              bpp of the image.                                           //
//  skip     = image to be skipped (default=-1)                             //
//  outcolor = output pixel value for image[]                               //
//  outbpp   = output bpp for image[]                                       //
//  screencolor = output pixel value for imz[] for display only (correc-    //
//    ted for grayvalue, and in screen bpp).                                //
//  noscreen = 1 if writing to screen is to be bypassed (write to image     //
//    buffer only). If noscreen is 1, the calling routine must call         //
//    rebuild_display() and copybackground() to draw the entire image.      //
//    This speeds things up when the image has to be re-palettized          //
//    anyway, and avoids putting "junk" pixels temporarily on screen.       //
//  win = 0 if setting pixel on drawing_area or on an image.  The actual    //
//    window is calculated automatically. If nonzero, it sets the pixel     //
//    on the specified window.                                              //
//  If ino is not 0, pixel is only drawn on specified image; otherwise,     //
//    image number to put pixel on is calc'd automatically.                 //
//  Returns the image no. on which the pixel was drawn.                     //
//--------------------------------------------------------------------------//
int setpixelonimage(int x, int y, uint color, int mode, int bpp, int skip, 
    int noscreen, Window win, int want_alpha, int ino)
{
   int xx=0,yy=0,cf,outbpp,outcolor,screencolor,color2, ubpp;
   ////  bpp = bits/pixel of color from calling routine for pixel
   ////  ubpp = bits/pixel of existing point underneath the pixel at x,y

   if(bpp==0) bpp=g.bitsperpixel;   
   if(g.inmenu)
   {  
          setpixel(x,y,color,mode,win);
   }else
   { 
      ////  If XOR'ing, convert underlying pixel to depth of the new 
      ////  color, then interact them at that depth.            

      if(mode!=SET && mode!=BUFFER)
      {   color2 = readpixelonimage(x,y,ubpp,ino,skip);  
          if(bpp!=ubpp) color2 = convertpixel(color2,ubpp,bpp,1);
          color = interact(color,color2,bpp,mode);
      }

      ////   Convert the interacted pixel to bpp of the topmost image. 

      if(ino==0) ino = whichimage(x,y,outbpp); else outbpp=bpp;
      if(mode==BUFFER)
      {   outcolor = color;
          screencolor = color;
      }else
      {   if(bpp==outbpp) outcolor=color;
          else            outcolor = convertpixel(color,bpp,outbpp,1);

          ////  Calculate color to put on screen.                      
          ////  If it is a monochrome image, remap the gray (grayvalue() also 
          ////   converts it to screen bpp).                         
          ////  If not drawing to screen, this can be skipped.    

          if(noscreen)
              screencolor=outcolor;
          else if(ino>=0 && z[ino].hitgray)
              screencolor=grayvalue(outcolor,ino,g.bitsperpixel);
          else
              screencolor=convertpixel(outcolor,outbpp,g.bitsperpixel,1);
      }
    
      ////  Set pixel in an image Window.                      
      ////  Put the modified pixel in the image[] or fft[] at outbpp bpp. 
      ////  Also put in spreadsheet if spreadsheet is visible.

      if(ino>=0) 
      {   xx = x - z[ino].xpos;
          yy = y - z[ino].ypos;
          cf = z[ino].cf;
          if(z[ino].floatexists) setfloatpixel(x,y,outcolor,ino);
          if(z[ino].waveletexists) setwaveletpixel(x,y,outcolor,ino);
          if(yy>=0 && yy<z[ino].ysize && xx>=0 && xx<z[ino].xsize)
          {   
              if(noscreen==0 && z[ino].s->visible) update_cell(ino,xx,yy);
              putpixelbytes(z[ino].image[cf][yy]+g.off[outbpp]*(xx),
                     outcolor, 1, outbpp);
              z[ino].touched=1;
          }
          if(want_alpha && g.alpha && mode!=BUFFER) set_alpha_bit(ino,cf,xx,yy,1); 
      }else
          if(want_alpha && g.alpha && mode!=BUFFER) set_alpha_bit(-1,0,x,y,1); 

      ////  If it's on the screen, set the pixel too. 
      ////  Use SET because color was already interacted.
      if(noscreen==0 && mode!=BUFFER) color = setpixel(x,y,screencolor,SET,win,ino);   
   }
   return ino;
}


//--------------------------------------------------------------------------//
//  setpixelonscreen                                                        //
//--------------------------------------------------------------------------//
void setpixelonscreen(int x, int y, int color)
{
    XSetForeground(g.display, g.image_gc, color); 
    XDrawPoint(g.display, g.main_window, g.image_gc, x, y);  
}


//--------------------------------------------------------------------------//
//  erase_pixel                                                             //
//--------------------------------------------------------------------------//
int erase_pixel(int x, int y, int *repairflags)
{
   int ii,iii,jj,ino,bpp,f,v,ct,done=0,needrepair=0,needremap=0;
   ino = whichimage(x,y,bpp);                     
   uchar *wadd, *radd;
   if(z[ino].backedup)
   {    f = z[ino].cf;
        ii = x - z[ino].xpos;
        iii = ii * g.off[bpp];
        jj = y - z[ino].ypos;
        if(between(ii,0,z[ino].xsize-1) && between(jj,0,z[ino].ysize-1))
        {    radd = z[ino].backup[f][jj]+iii;
             v = pixelat(radd, bpp);
             ct = z[ino].colortype;
             if(bpp>8 && ct==GRAY)    
                  v = grayvalue(v,ino,g.bitsperpixel);
             else if(bpp==8 && bpp!=g.bitsperpixel)
             {    needremap=1;
                  needrepair=1;
             }
             else if(bpp>8 && bpp!=g.bitsperpixel)
                  v = convertpixel(v,bpp,g.bitsperpixel,1);
             else if(g.bitsperpixel==8) 
             {    needremap=1;
                  needrepair=1;
             }
             wadd = z[ino].image[f][jj]+iii;
             putpixelbytes(wadd,v,1,bpp,1);
             if(needrepair==0 && needremap==0)
                  setpixelonimage(x,y,v,g.imode);
             else
             {   
                  v = convertpixel(v,bpp,g.bitsperpixel,1);
                  setpixel(x,y,v,SET);
             }
             set_alpha_bit(ino,f,x-z[ino].xpos,y-z[ino].ypos,0);
             z[ino].hit=1;
        }
   }else
   {    if(g.image_count>0) 
             message("Image was not backed up");
        else message("No images");
        done=1;
   } 
   if(done) repairflags[0] |= 4;
   if(needrepair) repairflags[ino] |= 2;
   if(needremap) repairflags[ino] |= 1;
   return ino;
}


//--------------------------------------------------------------------------//
//  set_alpha_bit                                                           //
//  x,y are indices in image buffer relative to upper left of image.        //
//--------------------------------------------------------------------------//
void set_alpha_bit(int ino, int frame, int x, int y, int value)
{
   int xindex, bit;
   xindex = x>>3;
   bit = 1<<(x-(xindex<<3));
   if(between(xindex,0,z[ino].xsize>>3-1) && between(y,0,z[ino].ysize-1))
   {   if(value) z[ino].alpha[frame][y][xindex] |= bit;
       else      z[ino].alpha[frame][y][xindex] &= ~bit;
   }
}


//--------------------------------------------------------------------------//
//  get_alpha_bit                                                           //
//  x,y are indices in image buffer relative to upper left of image.        //
//--------------------------------------------------------------------------//
int get_alpha_bit(int ino, int f, int x, int y)
{
   int xindex, bit, answer=0;
   xindex = x>>3;
   bit = 1<<(x-(xindex<<3));
   if(between(xindex,0,z[ino].xsize>>3-1) && between(y,0,z[ino].ysize-1))
      if(z[ino].alpha[f][y][xindex] & bit) answer=1;

   return answer;
}



//--------------------------------------------------------------------------//
//  zoom_x_coordinate                                                       //
//  Returns true x coordinate of a zoomed x coordinate                      //
//--------------------------------------------------------------------------//
int zoom_x_coordinate(int x, int ino)
{ 
  if(ino>=0 && z[ino].is_zoomed) 
     return(int)((double)(x - z[ino].xpos)*z[ino].zoomx_inv) + 
                          z[ino].zoom_x1 + z[ino].xpos;
  else return x;
}     

//--------------------------------------------------------------------------//
//  zoom_y_coordinate                                                       //
//  Returns true y coordinate of a zoomed y coordinate                      //
//--------------------------------------------------------------------------//
int zoom_y_coordinate(int y, int ino)
{ 
  if(ino>=0 && z[ino].is_zoomed) 
     return(int)((double)(y-z[ino].ypos)*z[ino].zoomy_inv)+z[ino].zoom_y1 + 
           z[ino].ypos;
  else return y;
}     


//--------------------------------------------------------------------------//
//  zoom_x_screen                                                           //
//  Returns screen x coordinate of a zoomed point given the unzoomed x      //
//  z[ino].zoom_xpos is distance from left of image where zoom starts       //
//--------------------------------------------------------------------------//
int zoom_x_screen(int x, int ino)
{ 
  if(ino>=0 && z[ino].is_zoomed) 
     return (int)((x - z[ino].zoom_x1 - z[ino].xpos) * z[ino].zoomx); 
  else return x;
}     

//--------------------------------------------------------------------------//
//  zoom_y_screen                                                           //
//  Returns screen y coordinate of a zoomed point given the unzoomed y      //
//--------------------------------------------------------------------------//
int zoom_y_screen(int y, int ino)
{ 
  if(ino>=0 && z[ino].is_zoomed) 
     return(int)((y - z[ino].zoom_y1 - z[ino].ypos) * z[ino].zoomy); 
  else return y;
}     


//--------------------------------------------------------------------------//
//  zoom_x_index                                                            //
//  Returns image buffer x index of a zoomed point given the zoomed x coord.//
//--------------------------------------------------------------------------//
int zoom_x_index(int x, int ino)
{ 
  if(ino>=0 && z[ino].is_zoomed) 
  return((int)(x*z[ino].zoomx_inv) + z[ino].zoom_x1); 
  else return x;
}     


//--------------------------------------------------------------------------//
//  zoom_y_index                                                            //
//  Returns image buffer y index of a zoomed point given the zoomed x coord.//
//--------------------------------------------------------------------------//
int zoom_y_index(int y, int ino)
{ 
  if(ino>=0 && z[ino].is_zoomed) 
  return((int)(y*z[ino].zoomy_inv) + z[ino].zoom_y1); 
  else return y;
}     


//--------------------------------------------------------------------------//
//  zoom_x_coordinate_of_index                                              //
//  Returns screen coordinate given the image buffer x index                //
//--------------------------------------------------------------------------//
int zoom_x_coordinate_of_index(int x, int ino)
{
   if(ino<0) return x;
   if(z[ino].is_zoomed) 
      return (int)((double)x*z[ino].zoomx_inv) + z[ino].zoom_x1 + z[ino].xpos;  
   return x+z[ino].xpos;
}


//--------------------------------------------------------------------------//
//  zoom_y_coordinate_of_index                                              //
//  Returns screen coordinate given the image buffer y index                //
//--------------------------------------------------------------------------//
int zoom_y_coordinate_of_index(int y, int ino)
{
   if(ino<0) return y;
   if(z[ino].is_zoomed) 
      return (int)((double)y*z[ino].zoomy_inv) + z[ino].zoom_y1 + z[ino].ypos;  
   return y+z[ino].ypos;
}


//--------------------------------------------------------------------------//
// interact - calculates the interaction of 2 pixels. They must have the    //
//  same bits/pixel.                                                        //
//--------------------------------------------------------------------------//
uint interact(int color1, int color2, int bpp, int mode)
{
     int rr,gg,bb,r2,g2,b2;
     if(bpp!=8 && bpp!=15 && bpp!=16 && bpp!=24 && bpp!=32) return 0;
     if(bpp==8)
     {  switch(mode)
        {   case MAXIMUM: color1=max(color1,color2); break;
            case MINIMUM: color1=min(color1,color2); break;     
            case ADD: color1=min((int)g.maxcolor,color1+color2); break;
            case SUB12: color1=max(0,color1-color2);     break;
            case SUB21: color1=max(0,color2-color1);     break;
            case XOR: color1^=color2; break;
            case AVE: color1+=color2; color1>>=1; break;       
            case SUP: if(color1==0) return(color2);
                      else return(color1);
        }
        color1=min(255,max(0,color1));
     }else
     {  valuetoRGB(color1,rr,gg,bb,bpp);
        valuetoRGB(color2,r2,g2,b2,bpp);
        switch(mode)
        {   case MAXIMUM:   rr=max(rr,r2); gg=max(gg,g2); bb=max(bb,b2); break;
            case MINIMUM:   rr=min(rr,r2); gg=min(gg,g2); bb=min(bb,b2); break;
            case ADD:   rr=min(g.maxred[bpp]  ,rr+r2);
                        gg=min(g.maxgreen[bpp],gg+g2);
                        bb=min(g.maxblue[bpp] ,bb+b2);
                        break;
            case SUB12: rr=max(0,rr-r2);
                        gg=max(0,gg-g2);
                        bb=max(0,bb-b2);
                        break;
            case SUB21: rr=max(0,r2-rr);
                        gg=max(0,g2-gg);
                        bb=max(0,b2-bb);
                        break;
            case XOR:   rr^=r2; gg^=g2; bb^=g2; 
                        break;
            case AVE:   rr+=r2; rr>>=1;
                        gg+=g2; gg>>=1;
                        bb+=b2; bb>>=1;
                        break;
            case SUP:   if(color1==0) return(color2); 
                        else return(color1);
        }
        color1 = RGBvalue(rr,gg,bb,bpp);
        color1=min((uint)g.maxvalue[bpp],(uint)max(0,color1));
     }   
     return((uint)color1);
}




//--------------------------------------------------------------------------//
// setfloatpixel - Converts an integer pixel value back to a floating       //
//  point number. This is done after manipulating a FFT-transformed         //
//  image. All manipulations are done on the integer value which is then    //
//  converted back to a float using scaling factors i_fmin[image]           //
//  and i_fmax[image].                                                      //
//  The calling routine has to know which image it is.                      //
//  x and y are screen coordinates of the pixel.                            //
//  If the fft'd image is showing "original image", it does nothing.        //
//--------------------------------------------------------------------------//
void setfloatpixel(int x, int y, uint color,int ino)
{
  complex value;
  double dreal=0.0,dimag=0.0;
  double dfac;
  if(ino>=0)
  { if( (z[ino].floatexists) && (x>=z[ino].xpos) && (y>=z[ino].ypos) &&
          (x<z[ino].xpos+z[ino].xsize) && (y<z[ino].ypos+z[ino].ysize)
          && (z[ino].fftdisplay!=NONE))
    {   x -= z[ino].xpos;
        y -= z[ino].ypos;
        dfac = (double)(z[ino].fmax-z[ino].fmin)/(double)g.maxvalue[z[ino].bpp]; 
        if(x>=0 && y>=0)
        {  
           switch(z[ino].fftdisplay)
           {  case REAL:  dimag=z[ino].fft[y][x].imag(); 
                          dreal = (double)color*dfac;
                          break;
              case IMAG:  dreal=real(z[ino].fft[y][x]); 
                          dimag = (double)color*dfac;
                          break;
              case POWR:  dreal = (double)color*dfac;
                          dimag = (double)color*dfac;
                          break;
              default:    fprintf(stderr,"error in setfloatpixel\n");
                          break;
           }   
           value.real()=dreal;
           value.imag()=dimag;
           z[ino].fft[y][x] = value;
        }
    }   
  }  
}


//--------------------------------------------------------------------------//
//whichimage - determines which image point (x,y) is on.                    //
//  Returns:      -1 = main window (not on an image)                        //
//                 0 to 511 = image number.                                 //
//  Also returns the bits/pixel of the image in ref. variable bpp.          //
//  The optional parameter `skip' allows one image to be skipped from       //
//     consideration for reading images in XOR mode.                        //
//  Images with the 'permanent' flag set to 0 are skipped.                  //
//  x and y are screen coordinates relative to (main_xpos, main_ypos).      //
//  Repeats until it finds an opaque pixel.                                 //
//--------------------------------------------------------------------------//
int whichimage(int x, int y, int& bpp, int skip)
{
   int cf, i, j, k, ino, tmax, tmin;
   int rr,gg,bb,rmin,rmax,gmin,gmax,bmin,bmax;
   uint v;
   for(k=0; k<g.image_count; k++)  
   {   ino = z[k].overlap;            // Start with foreground image
       if(z[ino].permanent==TEMP) continue;
       if(ino>0 && ino!=skip)
       {    if(insidebox(x,y,z[ino].xpos,z[ino].ypos, z[ino].xpos+z[ino].xsize-1,
                     z[ino].ypos+z[ino].ysize-1) )
            {    
                 bpp=z[ino].bpp;

                 ////  If pixel at that point is chromakeyed, skip it.
                 if(z[ino].chromakey)
                 {   if(z[ino].colortype==GRAY)
                     {     cf   = z[ino].cf;
                           tmin = z[ino].ck_graymin;
                           tmax = z[ino].ck_graymax;
                           i = x-z[ino].xpos;
                           j = y-z[ino].ypos;
                           v = pixelat(z[ino].image[cf][j] + i*g.off[bpp], bpp);  
                           if(!is_opaque(v,tmin,tmax,z[ino].chromakey)) continue; 
                     }else
                     {     cf   = z[ino].cf;
                           rmin = z[ino].ck_min.red;
                           rmax = z[ino].ck_max.red;
                           gmin = z[ino].ck_min.green;
                           gmax = z[ino].ck_max.green;
                           bmin = z[ino].ck_min.blue;
                           bmax = z[ino].ck_max.blue;
                           i = x-z[ino].xpos;
                           j = y-z[ino].ypos;
                           v = pixelat(z[ino].image[cf][j] + i*g.off[bpp], bpp);  
                           valuetoRGB(v,rr,gg,bb,bpp);
                           if(!is_opaque(rr,rmin,rmax,gg,gmin,gmax,bb,bmin,bmax,
                               z[ino].chromakey)) continue;
                     }
                 }
                 return(ino); 
            }   
       }
   }  
   bpp=g.bitsperpixel;
   return(0);
}
int whichimage(Widget w, int& bpp)
{
   int k;
   for(k=0; k<g.image_count; k++)  
   {     if(w==z[k].widget)
         {  bpp=z[k].bpp;
            return(k); 
         }
   }  
   bpp=g.bitsperpixel;
   return(0);
}
int whichimage(Window w, int& bpp)
{
   int k;
   for(k=0; k<g.image_count; k++)  
   {     if(w==z[k].win)
         {  bpp=z[k].bpp;
            return(k); 
         }
   }  
   bpp=g.bitsperpixel;
   return(0);
}



//--------------------------------------------------------------------------// 
// whatbpp - Returns the bits/pixel of whatever image is at x,y.            //
//--------------------------------------------------------------------------//
int whatbpp(int x, int y)
{
   int bpp;
   whichimage(x,y,bpp);
   return(bpp);
}


//--------------------------------------------------------------------------//
// image_number - obtain image no. from title                               //
// Strip off the "" if given.                                               //
//--------------------------------------------------------------------------//
int image_number(char *s)
{
   int k,start=0,end;
   static char *endptr=NULL;   
   static char tempstring[128];
   if(endptr==NULL) endptr = new char[128];
   if(s[0]=='\"') start=1;
   if(start) end = (int)index(s+1, '\"') - (int)s - start; // Position of final "
   else end=strlen(s);
   strcpy(tempstring,s+start);
   tempstring[end]=0;
   for(k=0;k<g.image_count;k++)  
       if(between(k,0,g.image_count-1) && !strcmp(basefilename(z[k].name), tempstring)) return k;
   //// If it's not a title, it must be the image number.
   strtol(tempstring, &endptr, 10);
   if(endptr[0])
   {     message("Invalid image name",ERROR); 
         g.getout=1; 
         return 0; 
   }
   return atoi(tempstring);
}


//--------------------------------------------------------------------------//
// double_image_number - obtain image no. as double for calculator.y        //
//--------------------------------------------------------------------------//
double double_image_number(char *s)
{
   double ii = (double)image_number(s);
   return ii;
}


//--------------------------------------------------------------------------//
// insidebox - returns 1 if x,y is inside box x1,y1,x2,y2                   //
//--------------------------------------------------------------------------//
int insidebox(int x, int y, int x1, int y1, int x2, int y2)
{
   if((x>=x1)&&(x<=x2)&&(y>=y1)&&(y<=y2)) return (1);
   else return(0);
}

//--------------------------------------------------------------------------//
// boxoverlap - returns 1 if 2 boxes overlap at any point                   //
//--------------------------------------------------------------------------//
int boxoverlap(int x1,int y1,int x2,int y2,int x3,int y3,int x4,int y4)
{  int y;
   for(y=y1;y<=y2;y++)
   {  if(between(y,y3,y4) && 
       ((between(x4,x1,x2)||between(x3,x1,x2))||
        (between(x1,x3,x4)||between(x2,x3,x4))))
      return(1);
   }
   return(0);
}


//--------------------------------------------------------------------------//
// readpixelonimage - Read color of a point in the selected current         //
//  image or on background if not within an image.                          //
//  x and y are virtual screen coordinates. Reads any bits/pixel.           //
//  Value returned is in the bits/pixel of the pixel, not necessarily       //
//  the screen. To get the screen equivalent, use readpixelonimg()          //
//  or readpixel().  To read a pixel from a Window not on drawing_area,     //
//  use readpixel().                                                        //
// Also returns the bits/pixel of the pixel in `bpp' and the image          //
//   no. in `ino'.                                                          //
// The optional parameter `skip' allows one image to be omitted from        //
//   consideration. This allows reading an image, then XOR'ing it with      //
//   the screen.                                                            //
// If the image no. is known with certainty, it is faster to call           //
//   pixelat() instead of this function.                                    //
//--------------------------------------------------------------------------//
uint readpixelonimage(int x, int y, int& bpp, int& ino, int skip)
{ 
   int wantimg=0;
   uint value=0;
   ino = whichimage(x,y,bpp,skip);
   int x1 = x-z[ino].xpos;
   int y1 = y-z[ino].ypos;  
   if(g.inmenu) bpp=g.bitsperpixel;
   uchar **array;
   if(ino>=0) 
   {  
      if(g.inmenu || z[ino].transparency) wantimg=1;
      if(wantimg) array = z[ino].img; 
      else array = z[ino].image[z[ino].cf];
      if(y >= z[ino].ypos && 
         y <  z[ino].ypos + z[ino].ysize && 
         x >= z[ino].xpos && 
         x <  z[ino].xpos + z[ino].xsize)
         {  if(z[ino].fftdisplay==NONE || wantimg)
               value = pixelat(array[y1] + x1*g.off[bpp], bpp);
            else
               value = pixel_equivalent(ino, x1, x1*g.off[bpp], y1); 
         }
   }
   return(value);   
}



//--------------------------------------------------------------------------//
// readpixel                                                                //
//--------------------------------------------------------------------------//
uint readpixel(int x, int y)
{
   int ino, bpp;
   uint value;
   g.inmenu++;
   value = readpixelonimage(x,y,bpp,ino,-2);
   g.inmenu--;
   return value;
}


//--------------------------------------------------------------------------//
// readpixelonimg                                                           //
//  Only 8,15,16,24, and 32 bpp screen modes are supported.                 //
//--------------------------------------------------------------------------//
uint readpixelonimg(int x, int y)
{ 
   int ino, bpp;
   ino = whichimage(x,y,bpp);
   if(ino>=0) return pixelat(z[ino].img[y-z[ino].ypos] + (x-z[ino].xpos)*g.off[bpp], bpp);
   return readpixel(x,y);                                     
}


//--------------------------------------------------------------------------//
// readbyte                                                                 //
// reads 8 pixels from screen, assumes B&W, compress into 1 byte            //
//--------------------------------------------------------------------------//
uchar readbyte(int x,int y)
{  
   uchar b;
   int k;
   uint threshold = g.maxcolor/2;
   b=0;
   for(k=0;k<8;k++) if(readpixel(x+k,y)>threshold) b |= 1<<(7-k);
   return(b);
}


//------------------------------------------------------------------------//
// itoa                                                                   //
// Convert integer to string.                                             //
//------------------------------------------------------------------------//
char *itoa(int value, char *str, int radix)
{
   switch(radix)
   {  
#ifdef MSDOS
      case 2:  sprintf(str, "%b",value); break;
#endif
      case 8:  sprintf(str, "%o",value); break; 
      case 10: sprintf(str, "%d",value); break; 
      case 16: sprintf(str, "%x",value); break; 
      default: perror("Error converting string"); exit(1);
   } 
   return(str);
}


//------------------------------------------------------------------------//
// ltoa                                                                   //
// Convert long integer to string.                                        //
//------------------------------------------------------------------------//
char *ltoa(long int value, char *str, int radix)
{
   return itoa((int)value, str, radix); 
}


//-----------------------------------------------------------------------//
//  gcvt                                                                 //
//  convert double to a string with signif significant digits.           //
//  no longer needed since linux now has a gcvt.                         //
//-----------------------------------------------------------------------//
#ifndef HAVE_GCVT
char *gcvt(double val, int signif, char *buf)
{  
   buf=buf; val=val;signif=signif;
   char gtemp[128];

#ifndef HAVE_FCVT
   sprintf(buf, "%f", val);
   strcpy(gtemp, buf);
   return buf;
#else
#ifdef HAVE_ISINF
   if(isinf(val)){ strcpy(buf,"Infinity");  return buf; }
#endif
#ifdef HAVE_ISNAN
   if(isnan(val)){ strcpy(buf,"Not a number"); return buf; }
#endif
   char *bufr;
   const int plussign=0;
   int b=0,j,k=0,position,sign,dotpos;

   bufr=fcvt(val,signif,&position,&sign);
   if(sign){ gtemp[0]='-'; k++; } 
   else if(plussign) {  gtemp[0]='+'; k++; }

   if(position>0)     // number is >= 1
   {   for(j=1;j<=position;j++,b++){ gtemp[k]=bufr[b]; k++; }
       gtemp[k]='.';
       k++;
   }    
   else               // number is < 1
   {  gtemp[k]='0';   // leading zero
      k++;
      gtemp[k]='.';
      k++;
      for(j=1;j<=-position;j++){ gtemp[k]='0'; k++; }
   }
   for(j=1;j<=signif;j++,b++){ gtemp[k]=bufr[b]; k++; }
  
   gtemp[k]=0;
         // double check to make sure excess zeros are not added
   dotpos = strlen(gtemp);
   for(k=0;k<(int)strlen(gtemp);k++) if(gtemp[k]=='.'){ dotpos=k;break; }
   if(signif>0)
     gtemp[min(128-1,dotpos+signif+1)]=0;
   else
     gtemp[min(128-1,dotpos+signif)]=0;
   strcpy(buf,gtemp);
   return buf;
#endif
}
#endif 


//-----------------------------------------------------------------------//
// maxgamma - find largest value in gamma table for an image             //
//-----------------------------------------------------------------------//
int maxgamma(int ino)
{
    static int oldino=0,oldresult=0;
    int k, highest=0;
    if(ino==oldino) return(oldresult);
    if(ino>=0)
    { 
      for(k=0;k<=255;k++) if(z[ino].gamma[k]>z[ino].gamma[highest]) highest=k;
      oldresult=z[ino].gamma[highest];
      oldino=ino;
      return(oldresult);
    }
    oldresult = (uint)g.maxvalue[g.bitsperpixel];
    return(oldresult);    
}


//--------------------------------------------------------------------------//
// intensity                                                                //
// Calculates the brightness of a pixel                                     //
//--------------------------------------------------------------------------//
int intensity(int pixelvalue, int bpp)
{
     int red, green, blue;
     valuetoRGB(pixelvalue, red, green, blue, bpp);
     return (int)(red*g.luminr + green*g.luming + blue*g.luminb);
}


//--------------------------------------------------------------------------//
// grayvalue                                                                //
//   calculates the gray-mapped value for a pixel if the image              //
//   is grayscale. If image is color, it has no effect.                     //
//   Each image has a different gray scale mapping, thus the                //
//   image no. is required.                                                 //
//   (Converts from real image value to screen-mapped value.)               //
//   `bpp' is desired bits/pixel of output, normally same as                //
//    screen bitsperpixel.                                                  //
//    Not to be confused with graypixel, which turns the value              //
//    into a "gray" color.                                                  //
//--------------------------------------------------------------------------//
int grayvalue(int value, int ino, int bpp)
{
  double fac=1.0;
  int rr,gg,bb;
  if(ino>=0)
  { 
     int whitein =   z[ino].gray[0];  // It is necessary to convert these to
     int blackin =   z[ino].gray[1];  // regular variables or it won't 
     int whiteout=   z[ino].gray[2];  // divide properly.
     int blackout=   z[ino].gray[3];         
     if(z[ino].hitgray)
     {  
        if((whitein!=whiteout)||(blackin!=blackout))
          if(whitein!=blackin)
            fac=(double)(whiteout-blackout)/(double)(whitein-blackin);
        value = (int)((value-blackin)*fac+blackout);
        value = max(0,min(whiteout,value));
   
        // Gray scale is mapped to the brightest single color in color modes,
        // so it is necessary to convert back to a real gray pixel value.

        if(bpp>8)                     
        { 
          value=max(0,min(value,g.maxred[bpp]));
          rr = value;
          gg = value; if(bpp==16) gg<<=1;  // 16bit mode has 6 bits of green
          bb = value;
          value = RGBvalue(rr,gg,bb,bpp);     
        }
     }   
  }   
  return(value);
}


//--------------------------------------------------------------------------//
// graypixel - Convert a pixel value at specified bpp to a grayscale        //
//            value, regardless of its bits/pixel. You can't just set       //
//            the r,g,and b the same because there may be different         //
//            bits for each color.                                          //
//            Used to convert a pixel value to a "gray" color before        //
//            displaying it, thus bpp is usually = bitsperpixel.            //
//                                                                          //
//            Not to be confused with grayvalue(), which remaps the         //
//            mapping of grayscale pixels to the screen.                    //
//--------------------------------------------------------------------------//
int graypixel(int value, int bpp)
{
   int rr,gg,bb;
   valuetoRGB(value,rr,gg,bb,bpp);
   switch(bpp)
   {   case 8: 
               return (int)((rr*4 + gg*4 + bb*4)/3);
       case 16:          
               bb=(rr+(gg>>1)+bb)/3;
               rr=bb;
               gg=bb*2;
               break;
       default:
               bb=(int)((double)rr*g.luminr +
                        (double)gg*g.luming +
                        (double)bb*g.luminb);
               bb = min(bb,g.maxblue[bpp]-1);
               rr=bb;
               gg=bb;
               break;
    }
   value = RGBvalue(rr,gg,bb,bpp);
   return value;
}     


//--------------------------------------------------------------------------//
//  graytocolor - convert a grayscale pixel to RGB values                   //
//--------------------------------------------------------------------------//
int graytocolor(int ino, int value, int bpp)
{
   int rr,gg,bb;
   double fac = (double)g.maxred[bpp] / (double)(z[ino].gray[0] - z[ino].gray[1]);
   value = cint((value - z[ino].gray[1]) * fac);
   rr=gg=bb=value;
   if(bpp==16) gg+=value;
   rr = max(0, min(rr, g.maxred[bpp]));
   gg = max(0, min(gg, g.maxgreen[bpp]));
   bb = max(0, min(bb, g.maxblue[bpp]));
   return RGBvalue(rr,gg,bb,bpp);
}


//--------------------------------------------------------------------------//
// cls - Clears the screen.                                                 //  
//--------------------------------------------------------------------------//
void cls(uint color)
{
   blackbox(1,0,g.main_xsize-2,g.main_ysize-2,color);
}



//--------------------------------------------------------------------------//
// null - do nothing funcs for clickbox(), plot(), etc.                     //
//--------------------------------------------------------------------------//
void null(int a)
{ a=a; }
void null(int a[10])
{ a[0]=a[0]; }
void null(double a)
{ a=a; }
void null(void)
{}
void null(pcb_struct v)
{ v.x=v.x; }
void null(dialoginfo *a, int b, int c, int d)
{ a=a; b=b; c=c; d=d;}
void null(void *p, int n1, int n2)
{   PlotData *pd = (PlotData *)p;
    pd->ymin=0;
    pd->ymax=0;
    n1=n1; n2=n2; 
}


//--------------------------------------------------------------------------//
// samplecolor - print sample of color in clickbox                          //
//--------------------------------------------------------------------------//
void samplecolor(int c)
{
  c=c;  // need more than this
}
void samplecolor(int c[10])
{
  c[0]=c[0];  // need more than this
}


//--------------------------------------------------------------------------//
// between    - returns 1 if a is between b and c                           //
//--------------------------------------------------------------------------//
int between(int a,int b,int c)
{   if((a>=b)&&(a<=c))return(1);
    else return(0);
}

//--------------------------------------------------------------------------//
// between    - returns 1 if a is between b and c                           //
//--------------------------------------------------------------------------//
int between(double a, double b, double c)
{   if((a>=b)&&(a<=c))return(1);
    else return(0);
}


//--------------------------------------------------------------------------//
// onedge - returns 1 if x,y is on edge of box x1,y1,x2,y2                  //
//--------------------------------------------------------------------------//
int onedge(int x, int y, int x1, int y1, int x2, int y2)
{
  const int w1 = 0;                     // width of handle outside box
  const int w2 = g.window_handle_size;  // width of handle inside box
  if( between(y, y1-w1, y2+w1))
  {  if( between(x, x1-w1, x1+w2)) return(1);
     if( between(x, x2-w2, x2+w1)) return(1);
  }
  if( between(x, x1-w1, x2+w1))
  {  if( between(y, y1-w1, y1+w2)) return(1);
     if( between(y, y2-w2, y2+w1)) return(1);
  }   
  return(0);
}


//--------------------------------------------------------------------------//
// stringtohex                                                              //
//  converts a string to integer assuming the string is a hex value.        //
//--------------------------------------------------------------------------//
int stringtohex(char* string)
{
     int k;
     char c;
     int answer = 0;
     for(k=0;k<(int)strlen(string);k++)
     {   answer <<= 4;
         c = string[k];
         if((c>='0')&&(c<='9')) answer += c-'0'; 
         if((c>='A')&&(c<='F')) answer += c-'A'+10; 
         if((c>='a')&&(c<='f')) answer += c-'a'+10; 
     }    
     return answer;
}


//--------------------------------------------------------------------------//
// savesettings                                                             //
//--------------------------------------------------------------------------//
void savesettings(void)
{
  FILE* fp;
  int k,rr,gg,bb;
  char *ifile;
  ifile = new char[FILENAMELENGTH];
  sprintf(ifile,"%s/tnimage.ini",g.homedir);

  setSVGApalette(-4, 0);     // Reset background palette
  valuetoRGB(g.fcolor,rr,gg,bb,g.bitsperpixel);
  g.fc.red=rr; g.fc.green=gg; g.fc.blue=bb;
  valuetoRGB(g.bcolor,rr,gg,bb,g.bitsperpixel);
  g.bc.red=rr; g.bc.green=gg; g.bc.blue=bb;

  if((fp=fopen(ifile, "wt"))!=NULL )
  { 
       fprintf(fp,"version %s\n",g.version);
       fprintf(fp,"auto_undo %d\n",g.autoundo);
       fprintf(fp,"background_color %d\n",g.bcolor);
       fprintf(fp,"background_palette %d\n",g.bkgpalette);
       fprintf(fp,"compression %s\n",g.compression);
       fprintf(fp,"decompression %s\n",g.decompression);
       fprintf(fp,"compression_ext %s\n",g.compression_ext);

       fprintf(fp,"default_font %s\n",g.font);
       fprintf(fp,"image_font %s\n",g.imagefont);
       fprintf(fp,"print_dither %d\n",printer.dither);
       fprintf(fp,"camera %s\n",g.camera);
       fprintf(fp,"camera_id %d\n",g.camera_id);
       fprintf(fp,"copy %d\n",g.copy);
       fprintf(fp,"double_click_msec %d\n",g.double_click_msec);
       fprintf(fp,"information_width %d\n",g.drawarea2width);
       fprintf(fp,"total_xsize %d\n",g.total_xsize);
       fprintf(fp,"total_ysize %d\n",g.total_ysize);
       fprintf(fp,"foreground_color %d\n",g.fcolor);
       fprintf(fp,"foreground_red %d\n",g.fc.red);
       fprintf(fp,"foreground_green %d\n",g.fc.green);
       fprintf(fp,"foreground_blue %d\n",g.fc.blue);
       fprintf(fp,"background_red %d\n",g.bc.red);
       fprintf(fp,"background_green %d\n",g.bc.green);
       fprintf(fp,"background_blue %d\n",g.bc.blue);
       fprintf(fp,"help_directory %s\n",g.helpdir);
       fprintf(fp,"format_path %s\n",g.formatpath);
       fprintf(fp,"jpeg_q_factor %d\n",g.jpeg_qfac);
       fprintf(fp,"luminosity_red %g\n",g.luminr);
       fprintf(fp,"luminosity_green %g\n",g.luming);
       fprintf(fp,"luminosity_blue %g\n",g.luminb);
       fprintf(fp,"object_threshold_red %d\n",g.object_threshold.red);
       fprintf(fp,"object_threshold_green %d\n",g.object_threshold.green);
       fprintf(fp,"object_threshold_blue %d\n",g.object_threshold.blue);

       fprintf(fp,"print_paper_xsize %g\n",printer.paperxsize);
       fprintf(fp,"print_paper_ysize %g\n",printer.paperysize);
       fprintf(fp,"print_blue_factor %g\n",printer.bfac); 
       fprintf(fp,"print_depletion %d\n",printer.depletion);
       fprintf(fp,"print_graybalance %d\n",printer.graybalance);
       fprintf(fp,"print_green_factor %g\n",printer.gfac);
       fprintf(fp,"print_intensity_factor %g\n",printer.ifac);
       fprintf(fp,"print_palette %d\n",printer.palette);
       fprintf(fp,"print_paper %d\n",printer.papertype);
       fprintf(fp,"print_language %d\n",printer.language);
       fprintf(fp,"print_positive %d\n",printer.positive);
       fprintf(fp,"print_quality %d\n",printer.quality);
       fprintf(fp,"print_red_factor %g\n",printer.rfac);
       fprintf(fp,"print_resolution %d\n",printer.res);
       fprintf(fp,"print_device %d\n",printer.device);
       fprintf(fp,"print_vertical %d\n",printer.vertical);
       fprintf(fp,"print_hsize %g\n",printer.hsize);
       fprintf(fp,"print_ratio %g\n",printer.ratio);
       fprintf(fp,"print_rotation %d\n",printer.rotation);
       fprintf(fp,"print_interpolate %d\n",printer.interpolate);
       fprintf(fp,"print_command %s\n",printer.command);
       fprintf(fp,"print_devicename %s\n",printer.devicename);
       fprintf(fp,"print_hpos %g\n",printer.hpos);
       fprintf(fp,"print_vpos %g\n",printer.vpos);

       fprintf(fp,"read_cmyk %d\n",g.read_cmyk);
       fprintf(fp,"save_cmyk %d\n",g.save_cmyk);
       fprintf(fp,"scanner %s\n",g.scan_device);
       fprintf(fp,"scannerid %d\n",g.scannerid);
       fprintf(fp,"signif_digits %d\n",g.signif);
       fprintf(fp,"spacing %d\n",g.spacing);
       fprintf(fp,"text_spacing %d\n",g.text_spacing);
       fprintf(fp,"use_custom_header %d\n",g.useheader);
       fprintf(fp,"want_colormaps %d\n",g.want_colormaps);
       fprintf(fp,"want_crosshairs %d\n",g.want_crosshairs);
       fprintf(fp,"want_messages %d\n",g.want_messages);
       fprintf(fp,"want_quantize %d\n",g.want_quantize);
       fprintf(fp,"want_raise %d\n",g.want_raise);
       fprintf(fp,"want_shell %d\n",g.want_shell);
       fprintf(fp,"want_slew %d\n",g.want_slew);
       fprintf(fp,"want_byte_order %d\n",g.want_byte_order);
       fprintf(fp,"wavelet_path %s\n",g.waveletpath);
       fprintf(fp,"window_border %d\n",g.window_border);
       fprintf(fp,"window_handle_size %d\n",g.window_handle_size);
       fprintf(fp,"xlib_cursor %d\n",g.xlib_cursor);
       fprintf(fp,"nbuttons %d\n",g.nbuttons);
       for(k=0;k<g.nbuttons;k++)
          fprintf(fp,"button %s %s:%s\n",
                  g.button[k]->label, g.button[k]->activate_cmd,
                  g.button[k]->deactivate_cmd);
       fclose(fp);

  }else
  {    fprintf(stderr,"Can't save settings\n");
       error_message(ifile, errno);
  }
  
  //// Make sure user can still read ini file even if tnimage was run as root.
  chmod(ifile, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
  delete[] ifile;
}


//--------------------------------------------------------------------------//
// readsettings                                                             //
//--------------------------------------------------------------------------//
void readsettings(void)
{  
  if(g.diagnose){ printf("Reading tnimage.ini...");fflush(stdout); }
  FILE* fp;
  int gotconfig=1, status=0;
  int space=0, ret=0, but=0,k, length;
  char name[FILENAMELENGTH]="";
  char value[FILENAMELENGTH]="";
  char initfile[FILENAMELENGTH]="";
  char initversion[128]="";
  char tempstring[FILENAMELENGTH]="";
  int colon=0;

  sprintf(initfile,"%s/tnimage.ini",g.homedir);
  if((fp=fopen(initfile,"rb"))==NULL ) gotconfig=0;
  else fclose(fp);

  initversion[0]=0;
  if(gotconfig)
  {  fp=fopen(initfile,"rt");
     while(!feof(fp))
     { 
       fgets(name, 80, fp);
       if(feof(fp))break;
       space = strcspn(name," ");                // Position of 1st space
       ret = (int)index(name, '\n') - (int)name; // Position of final \n
       name[space]=0;
       name[ret]=0;
       strcpy(value, 1+name+space);
       if(strcmp(name,"version")==SAME) strcpy(initversion,value); 
       if(strcmp(initversion, g.version)!=SAME)
       {   fprintf(stderr,"Version mismatch in tnimage.ini - ignoring old settings\n");
           gotconfig=0;
           break;
       }

       ///-----parameters-----///
       if(strcmp(name,"background_palette")==SAME) g.bkgpalette=atoi(value);
       if(g.bkgpalette<=0) g.bkgpalette=1;
       if(strcmp(name,"information_width")==SAME) g.drawarea2width=atoi(value);
       if(strcmp(name,"total_xsize")==SAME)
       {   g.total_xsize = atoi(value);
           g.main_xsize = g.total_xsize-g.drawarea2width;   
       }
       if(strcmp(name,"total_ysize")==SAME) 
       {   g.total_ysize = atoi(value);
           g.main_ysize = g.total_ysize+(int)g.mainmenuheight;
       }
       if(strcmp(name,"button")==SAME) 
       {   if(but>=g.nbuttons)
               fprintf(stderr,"tnimage.ini: Error: button count mismatch, ignoring user button\n");
           else
           {   
               space = strcspn(value," ");           // Position of 1st space
               colon = strcspn(value,":");           // Position of 1st colon
               strncpy(g.button[but]->label, value, space);
               g.button[but]->label[space]=0;
               strncpy(g.button[but]->activate_cmd, value+space+1, colon-space-1);
               g.button[but]->activate_cmd[colon-space-1]=0;
               if(colon<(int)strlen(value))
                  strncpy(g.button[but]->deactivate_cmd, value+colon+1, COMMAND_LENGTH);
               else
                  strcpy(g.button[but]->deactivate_cmd, "null");
               but++;
           }
       }

       if(strcmp(name,"camera")==SAME) strcpy(g.camera, value);
       if(strcmp(name,"camera_id")==SAME) g.camera_id=atoi(value);
       if(strcmp(name,"compression")==SAME) strcpy(g.compression, value);
       if(strcmp(name,"decompression")==SAME) strcpy(g.decompression, value);
       if(strcmp(name,"compression_ext")==SAME) strcpy(g.compression_ext, value);
       if(strcmp(name,"default_font")==SAME) strcpy(g.font, value);
       if(strcmp(name,"image_font")==SAME) strcpy(g.imagefont, value);
       if(strcmp(name,"foreground_color")==SAME) g.fcolor=atoi(value);
       if(strcmp(name,"background_color")==SAME) g.bcolor=atoi(value);
       if(strcmp(name,"double_click_msec")==SAME) g.double_click_msec=atoi(value);
       if(strcmp(name,"foreground_red")==SAME) g.fc.red=atoi(value);
       if(strcmp(name,"foreground_green")==SAME) g.fc.green=atoi(value);
       if(strcmp(name,"foreground_blue")==SAME) g.fc.blue=atoi(value);
       if(strcmp(name,"background_red")==SAME) g.bc.red=atoi(value);
       if(strcmp(name,"background_green")==SAME) g.bc.green=atoi(value);
       if(strcmp(name,"background_blue")==SAME) g.bc.blue=atoi(value);
       if(strcmp(name,"format_path")==SAME)
       {   strcpy(g.formatpath, value);
           length = strlen(g.formatpath);
           if(length && g.formatpath[length] =='/') g.formatpath[length]=0;
       }
       if(strcmp(name,"help_directory")==SAME) strcpy(g.helpdir, value);
       if(strcmp(name,"luminosity_red")==SAME) g.luminr=atof(value);
       if(strcmp(name,"luminosity_green")==SAME) g.luming=atof(value);
       if(strcmp(name,"luminosity_blue")==SAME) g.luminb=atof(value);
       if(strcmp(name,"copy")==SAME) g.copy=atoi(value);
       if(strcmp(name,"nbuttons")==SAME) 
       {   g.nbuttons = max(0,atoi(value));
           g.button = new ButtonStruct*[g.nbuttons];
           for(k=0;k<g.nbuttons;k++)
           {   g.button[k] = new ButtonStruct;
               g.button[k]->button = 0;
               g.button[k]->state  = OFF;
               g.button[k]->widget = 0;
               g.button[k]->label  = new char[LABEL_LENGTH];
               g.button[k]->activate_cmd = new char[COMMAND_LENGTH];
               g.button[k]->deactivate_cmd = new char[COMMAND_LENGTH];
           }
       }
       if(strcmp(name,"nosparse")==SAME) g.nosparse=atoi(value);
       if(strcmp(name,"object_threshold_red")==SAME) g.object_threshold.red=atoi(value);
       if(strcmp(name,"object_threshold_green")==SAME) g.object_threshold.green=atoi(value);
       if(strcmp(name,"object_threshold_blue")==SAME) g.object_threshold.blue=atoi(value);

       if(strcmp(name,"print_quality")==SAME)   printer.quality=atoi(value);
       if(strcmp(name,"print_paper_xsize")==SAME)   printer.paperxsize=atof(value);
       if(strcmp(name,"print_paper_ysize")==SAME)   printer.paperysize=atof(value);
       if(strcmp(name,"print_device")==SAME)    printer.device=atoi(value);
       if(strcmp(name,"print_dither")==SAME)    printer.dither=atoi(value);
       if(strcmp(name,"print_resolution")==SAME)printer.res=atoi(value);
       if(strcmp(name,"print_vertical")==SAME)  printer.vertical=atoi(value);
       if(strcmp(name,"print_positive")==SAME)  printer.positive=atoi(value);
       if(strcmp(name,"print_palette")==SAME)   printer.palette=atoi(value);
       if(strcmp(name,"print_language")==SAME)  printer.language=atoi(value);
       if(strcmp(name,"print_intensity_factor")==SAME) printer.ifac=atof(value);
       if(strcmp(name,"print_red_factor")==SAME) printer.rfac=atof(value);
       if(strcmp(name,"print_green_factor")==SAME) printer.gfac=atof(value);
       if(strcmp(name,"print_blue_factor")==SAME) printer.bfac=atof(value);
       if(strcmp(name,"print_paper")==SAME) printer.papertype=atoi(value);
       if(strcmp(name,"print_hsize")==SAME) printer.hsize=atof(value);
       if(strcmp(name,"print_ratio")==SAME) printer.ratio=atof(value);
       if(strcmp(name,"print_rotation")==SAME) printer.rotation=atoi(value);
       if(strcmp(name,"print_interpolate")==SAME) printer.interpolate=atoi(value);
       if(strcmp(name,"print_hpos")==SAME) printer.hpos=atof(value);
       if(strcmp(name,"print_vpos")==SAME) printer.vpos=atof(value);

       if(strcmp(name,"read_cmyk")==SAME) g.read_cmyk=atoi(value);
       if(strcmp(name,"auto_undo")==SAME) g.autoundo=atoi(value);
       if(strcmp(name,"use_custom_header")==SAME) g.useheader=atoi(value);
       if(strcmp(name,"save_cmyk")==SAME) g.save_cmyk=atoi(value);
       if(strcmp(name,"signif_digits")==SAME) g.signif=min(8,atoi(value));
       if(strcmp(name,"sparse")==SAME) g.sparse=atoi(value);
       if(strcmp(name,"print_graybalance")==SAME) printer.graybalance=atoi(value);
       if(strcmp(name,"print_depletion")==SAME) printer.depletion=atoi(value);
       if(strcmp(name,"jpeg_q_factor")==SAME) g.jpeg_qfac=atoi(value);
       if(strcmp(name,"print_command")==SAME) strcpy(printer.command,value);
       if(strcmp(name,"print_devicename")==SAME) strcpy(printer.devicename,value);
       if(strcmp(name,"scanner")==SAME) strcpy(g.scan_device,value);
       if(strcmp(name,"scannerid")==SAME) g.scannerid=atoi(value);
       if(strcmp(name,"spacing")==SAME) g.spacing=atoi(value);
       if(strcmp(name,"text_spacing")==SAME) g.text_spacing=atoi(value);
       if(strcmp(name,"want_colormaps")==SAME)
       {      g.want_colormaps=atoi(value);
              if(g.want_colormaps==2) g.want_colormaps=0; // < 32 colors ast time
       }
       if(strcmp(name,"want_crosshairs")==SAME) g.want_crosshairs=atoi(value);
       if(strcmp(name,"want_messages")==SAME) g.want_messages=atoi(value);
       if(strcmp(name,"want_quantize")==SAME) g.want_quantize=atoi(value);
       if(strcmp(name,"want_slew")==SAME) g.want_slew=atoi(value);
       if(strcmp(name,"want_raise")==SAME) g.want_raise=atoi(value);
       if(strcmp(name,"want_shell")==SAME) g.want_shell=atoi(value);
       if(strcmp(name,"want_byte_order")==SAME) g.want_byte_order=atoi(value);
       if(strcmp(name,"wavelet_path")==SAME)
       {   strcpy(g.waveletpath, value);
           length = strlen(g.waveletpath);
           if(length && g.waveletpath[length] =='/') g.waveletpath[length]=0;
       }
       if(strcmp(name,"window_border")==SAME) g.window_border=atoi(value);
       if(strcmp(name,"window_handle_size")==SAME) g.window_handle_size=atoi(value);
       if(strcmp(name,"xlib_cursor")==SAME) g.xlib_cursor=atoi(value);
     }
     fclose(fp);
     
  }
  if(!gotconfig)
  {
       if(g.diagnose) printf("access homedir %d\n",access(g.homedir,F_OK));
       if(access(g.homedir,F_OK)) 
       {    printf("\nCreating directory %s\n", g.homedir);        
            status = mkdir(g.homedir, 0755);
            if(status) printf("Can't create directory:\n%s",g.homedir);

            sprintf(tempstring,"%s/formats",g.homedir);
            status = mkdir(tempstring, 0755);
            if(status) printf("Can't create directory:\n%s",tempstring);
       }
       if(g.diagnose) printf("status %d\n",status);       
       if(g.diagnose){ printf("%s/tnimage.ini not found, setting default buttons\n",
             g.homedir);fflush(stdout);}
       const int NBUTTONS=20;
       char buttonname[NBUTTONS][20] ={ 
                                        "Cancel",           "Slew",
                                        "Open",             "Close",
                                        "New",              "Save", 
                                        "Sketch",           "Line", 
                                        "Copy",             "Erase", 
                                        "Scan",             "Print", 
                                        "Label",            "Zoom",
                                        "Fcolor",           "Bcolor",
                                        "Densit",           "Config",
                                        "Prop",             "Quit"
                                      };
       char activatecmd[NBUTTONS][20]  ={ 
                                        "cancel",           "slew",
                                        "open",             "close",
                                        "new",              "save", 
                                        "sketch",           "line", 
                                        "move",             "erase", 
                                        "scan",             "print", 
                                        "label",            "zoom",
                                        "foregroundcolor",  "backgroundcolor",
                                        "stripdensitometry","configure",
                                        "attributes",       "quit", 
                                      };
       char deactivatecmd[NBUTTONS][20]  ={ 
                                        "null",             "slew",
                                        "null",             "null",
                                        "null",             "null", 
                                        "cancel",           "cancel", 
                                        "move",             "null", 
                                        "null",             "null", 
                                        "null",             "repair",
                                        "null",             "null",
                                        "null",             "null",
                                        "null",             "null"
                                      };
       g.nbuttons = NBUTTONS;
       g.button = new ButtonStruct*[NBUTTONS];
       for(k=0;k<NBUTTONS;k++) 
       {   g.button[k] = new ButtonStruct;
           g.button[k]->button = 0;
           g.button[k]->state  = OFF;
           g.button[k]->widget = 0;
           g.button[k]->label = new char[LABEL_LENGTH];
           g.button[k]->activate_cmd = new char[COMMAND_LENGTH];
           g.button[k]->deactivate_cmd = new char[COMMAND_LENGTH];
           strcpy(g.button[k]->label, buttonname[k]);
           strcpy(g.button[k]->activate_cmd, activatecmd[k]);
           strcpy(g.button[k]->deactivate_cmd, deactivatecmd[k]);
       }
  }
  g.currentpalette=g.bkgpalette;
  if(g.diagnose)
  {    printf("done\n nbuttons=%d\n",g.nbuttons);
       for(k=0;k<g.nbuttons;k++)
           printf(" button %d %s %s\n",k,g.button[k]->label,
               g.button[k]->activate_cmd);
       print_configuration();
  }
}        



//--------------------------------------------------------------------------//
// basefilename                                                             //
// strip path from filename & returns file name w/o path                    //
//--------------------------------------------------------------------------//
char* basefilename(char* pathname)
{  int k;
   int slashpos=0;
   char temp[FILENAMELENGTH]="";
   if(strlen(pathname)) strcpy(temp,pathname);

   for(k=0;k<(int)strlen(temp);k++)          // find last slash or colon
       if((temp[k]=='/')||(temp[k]=='\\')||(temp[k]==':')) slashpos=k+1;
   return(temp+slashpos);
}


//--------------------------------------------------------------------------//
// strip_relative_path                                                      //
// strip path from filename only if it is a relative path                   //
//--------------------------------------------------------------------------//
char* strip_relative_path(char* pathname)
{  int k;
   int slashpos=0;
   char temp[FILENAMELENGTH]="";
   if(strlen(pathname)) strcpy(temp,pathname);

   if(temp[0]=='.' && (temp[1]=='/') || (temp[1]=='.' && temp[2]=='/'))
   {   for(k=0;k<(int)strlen(temp);k++)          // find last slash or colon
           if((temp[k]=='/')||(temp[k]=='\\')||(temp[k]==':')) slashpos=k+1;
   }
   return(temp+slashpos);
}



//--------------------------------------------------------------------------//
// print_configuration                                                      //
//--------------------------------------------------------------------------//
void print_configuration(void)
{   int k;
    printf("Configuration\n"); 
    printf(" cursor       %d\n",g.xlib_cursor); 
    printf(" quantize     %d\n",g.want_quantize); 
    printf(" autoundo     %d\n",g.autoundo); 
    if(g.bitsperpixel==8) 
    {   printf(" reserved colors\n"); 
        for(k=0;k<256;k++)if(g.reserved[k])printf("%d ",k); 
        printf("\n");
    }
    printf(" luminosity   %g %g %g\n",g.luminr,g.luming,g.luminb); 
    printf(" colormaps    %d\n",g.want_colormaps); 
    printf(" signif       %d\n",g.signif); 
    printf(" csize        %d\n",g.csize); 
    printf(" sprayfac     %d\n",g.sprayfac); 
    printf(" total xsize  %d\n",g.total_xsize);
    printf(" total ysize  %d\n",g.total_ysize);
    printf(" main xsize   %d\n",g.main_xsize);
    printf(" main ysize   %d\n",g.main_ysize);
    printf(" xlib cursor  %d\n",g.xlib_cursor);
    printf(" fcolor       %d\n",g.fcolor);
    printf(" bcolor       %d\n",g.bcolor);
    printf(" default font %s\n",g.font);
    printf(" image font   %s\n",g.imagefont);
    printf(" infor. width %d\n",g.drawarea2width);
    printf(" mainmenuheight %d\n",(int)g.mainmenuheight);
    printf(" slew         %d\n",g.want_slew);
    printf(" raise        %d\n",g.want_raise);
    printf(" shell        %d\n",g.want_shell);
    printf(" colortype    %d\n",g.colortype);
    printf(" draw_figure  %d\n",g.draw_figure);
    printf(" window_border %d\n",g.window_border);
    printf(" window_handle size %d\n",g.window_handle_size);
    printf(" home dir     %s\n",g.homedir);
    printf(" start dir    %s\n",g.startdir);
    printf(" current dir  %s\n",g.currentdir);
    printf(" help dir     %s\n",g.helpdir);
    fflush(stdout);
}


//--------------------------------------------------------------------------//
// print_image_info                                                         //
//--------------------------------------------------------------------------//
void print_image_info(void)
{   int k;
    printf("Image information\n"); 
    for(k=0;k<g.image_count;k++)
    {  
       printf("ino %d x=%d y=%d bpp=%d\n",k,z[k].xsize,z[k].ysize,z[k].bpp);
       printf("   xpos  %d ypos %d frames %d %d\n",z[k].xpos,z[k].ypos,z[k].frames,z[k].cf);
       printf("   transparency %d chromakey %d %d %d %d %d %d\n",
          z[k].transparency,
          z[k].chromakey, 
          z[k].ck_graymin,
          z[k].ck_graymax,
          z[k].ck_min.red,
          z[k].ck_min.green, 
          z[k].ck_min.blue);
       printf("   xpos  %d ypos %d \n",z[k].xpos,z[k].ypos);
       printf("   backed up %d \n",z[k].backedup);
       printf("   overlap   %d \n",z[k].overlap);
       printf("   colortype %d \n",z[k].colortype);
       printf("   fft %d %d %g %g %d %d %d\n",
          z[k].floatexists,
          z[k].fftdisplay,
          z[k].fmin,
          z[k].fmax,
          z[k].fftxsize,
          z[k].fftysize,
          z[k].fftstate);
       printf("   shell %d %d %d \n",z[k].shell,z[k].permanent,z[k].window_border);
    }
}    



//--------------------------------------------------------------------------//
// is_opaque                                                                //
//--------------------------------------------------------------------------//
int is_opaque(int v, int tmin, int tmax, int chromakey)
{
      switch(chromakey)
      {    case CHROMAKEY_NONE:   return 1;
           case CHROMAKEY_EQUAL:  if(abs(v-tmin)>5) return 1;
                                  else return 0;
           case CHROMAKEY_NOTEQUAL:if(abs(v-tmin)<5) return 1;
                                  else return 0;
           case CHROMAKEY_NORMAL: if(between(v,tmin,tmax)) return 1;
                                  else return 0;
           case CHROMAKEY_INVERT: if(!between(v,tmin,tmax)) return 1;
                                  else return 0;
           default: return 1;
      }
}
int is_opaque(int rr, int rmin, int rmax,
              int gg, int gmin, int gmax,
              int bb, int bmin, int bmax, int chromakey)
{
      switch(chromakey)
      {    case CHROMAKEY_NONE:  return 1;
           case CHROMAKEY_EQUAL: if(abs(rr-rmin)>5 ||
                                    abs(gg-gmin)>5 ||
                                    abs(bb-bmin)>5) return 1;
                                 else return 0;
           case CHROMAKEY_NOTEQUAL: if(abs(rr-rmin)<5 ||
                                    abs(gg-gmin)<5 ||
                                    abs(bb-bmin)<5) return 1;
                                 else return 0;
           case CHROMAKEY_NORMAL:if(between(rr,rmin,rmax) &&
                                    between(gg,gmin,gmax) &&
                                    between(bb,bmin,bmax)) return 1;
                                 else return 0;
           case CHROMAKEY_INVERT:if(!between(rr,rmin,rmax) &&
                                    !between(gg,gmin,gmax) &&
                                    !between(bb,bmin,bmax)) return 1;
                                 else return 0;
           default: return 1;
      }
}


//--------------------------------------------------------------------------//
// root - put image on root window                                          //
// mode can be ON OFF PERMANENT_TILE                                        //
//--------------------------------------------------------------------------//
void root(int ino, int mode)
{
   int j,k,bpp,width,height;
   if(ino<=0){ message("Please select an image"); return; }
   Window owin = z[ino].win;
   Pixmap pixmap;
   int imgxsize = z[ino].xsize;
   int imgysize = z[ino].ysize;
   int xcount = (int)(0.9999 + (double)g.xres / (double)imgxsize);
   int ycount = (int)(0.9999 + (double)g.yres / (double)imgysize);
   XImage *image;
   switch(mode)
   {   case ON:  
           z[ino].win = g.root_window; 
           break;
       case PERMANENT_TILE:  
           ////  Don't let X11 tile it automatically, it will screw up by spacing out
           ////  the image to fit screen evenly, leaving black areas.
           bpp = g.bitsperpixel; 
           if(bpp<8) bpp=8;
           if(g.sparse_packing && bpp==32) bpp=24;
           pixmap = XCreatePixmap(g.display, g.root_window, g.xres, g.yres, bpp);
           image = createximage(imgxsize, imgysize, bpp, z[ino].img_1d);
           for(j=0;j<ycount;j++)
           for(k=0;k<xcount;k++)
           {    width = min(imgxsize, g.xres-k*imgxsize);
                height = min(imgysize, g.yres-j*imgysize);
                XPutImage(g.display, pixmap, g.image_gc, image, 0, 0, 
                     k*imgxsize, j*imgysize, width, height); 
                copyimage(0,0,width,height,k*imgxsize,j*imgysize,PUT,image,g.root_window);
           }
           XSetWindowBackgroundPixmap(g.display, g.root_window, pixmap);
           XFreePixmap(g.display, pixmap);
           send_event(g.root_window,Expose,0,0,g.xres,g.yres);
           break;
       case OFF: 
           z[ino].win = XtWindow(z[ino].widget); 
           break;
   }
   redraw(ino);  
   redrawscreen();
   copybackground(0,0,z[ino].xsize,z[ino].ysize,-1);
   send_event(z[ino].win,Expose,0,0,z[ino].xsize,z[ino].ysize);
   send_event(z[ino].win,NoExpose,0,0,z[ino].xsize,z[ino].ysize);
   send_event(owin,Expose,0,0,z[ino].xsize,z[ino].ysize);
}


//--------------------------------------------------------------------------//
// changetitle                                                              //
//--------------------------------------------------------------------------//
void changetitle(int noofargs, char **arg)
{ 
   int ino=ci;
   if(noofargs>=1) ino=image_number(arg[1]); else ino=ci;
   if(g.getout){ g.getout=0; return; }
   if(!between(ino,0,g.image_count-1)) { message("Bad image number",ERROR); return; }
   if(noofargs>=2) strncpy(z[ino].name, arg[2], FILENAMELENGTH-1);
   else message("New image name:",z[ino].name,PROMPT,FILENAMELENGTH-1);
   setimagetitle(ino, z[ino].name);
}


//--------------------------------------------------------------------------//
// string_index - does what index() should be doing: returns the position   //
// of the 1st or lastt occurrence of c in text.   If not found, returns 0.  //
//--------------------------------------------------------------------------//
int string_index(char *text, int direction, int c)
{
    int k;
    int len=strlen(text);
    if(direction>0)
    {    for(k=0;k<len;k++) if(text[k]==c) return k; }
    else
    {    for(k=len-1;k>=0;k--) if(text[k]==c) return k; }
    return 0;
}


